# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

import subprocess

__all__ = ["GitStateGetter"]


class GitStateGetter:
    def __init__(self) -> None:
        pass

    @staticmethod
    def _git_command(command: str) -> str:
        process = subprocess.Popen(
            command.split(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        stdout, stderr = process.communicate()
        return str(stdout)

    def create_git_string(self) -> str:
        self.string = ""

        self.string += "\n"
        self.string += "----- REMOTE ------ \n"
        self.string += self._git_command("git remote -v")

        self.string += "\n"
        self.string += "----- BRANCH ------ \n"
        self.string += self._git_command("git branch -v")

        self.string += "\n"
        self.string += "----- LAST COMMIT ------ \n"
        self.string += self._git_command("git rev-parse HEAD")

        self.string += "\n"
        self.string += "----- STATUS ------ \n"
        self.string += self._git_command("git status")

        self.string += "\n"
        self.string += "----- DIFFERENCE ------ \n"
        self.string += self._git_command("git diff HEAD")

        return self.string



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2024 Flavio Lorez and contributors
#
# There is no warranty for this code

from __future__ import annotations

import os
import subprocess
from functools import wraps

from bamboost.common.utilities import to_camel_case
from bamboost.simulation_writer import SimulationWriter

__all__ = ["install"]


NAME_OF_DICT = "_slurm"


def _extend_enter_slurm_info(original_enter):
    @wraps(original_enter)
    def modified_enter(self: SimulationWriter, *args, **kwargs):
        slurm_job_id = os.environ.get("SLURM_JOB_ID")
        self.update_metadata({NAME_OF_DICT: {"jobId": slurm_job_id}})
        return original_enter(self, *args, **kwargs)

    return modified_enter


def _extend_exit_slurm_info(original_exit):
    @wraps(original_exit)
    def modified_exit(self: SimulationWriter, exc_type, exc_value, exc_tb):
        # Inject the following into the __exit__ method of SimulationWriter
        slurm_job_id = os.environ.get("SLURM_JOB_ID")

        result = subprocess.run(
            ["myjobs", "-j", slurm_job_id], env=os.environ, capture_output=True
        )

        # Check if the command was successful
        if result.returncode == 0:
            # Get the output as a string
            output_str = result.stdout.decode("utf-8")

            # Create a dictionary to store the key-value pairs
            slurm_dict = {}
            for line in output_str.strip().split("\n"):
                if ":" in line:  # Ensure the line contains a key-value pair
                    key, value = line.split(":", 1)  # Split on the first colon

                    # modify key to camelCase
                    slurm_dict[to_camel_case(key.strip())] = value.strip()

            # _write_slurm_info(self, slurm_dict)
            self.update_metadata({NAME_OF_DICT: slurm_dict})

        return original_exit(self, exc_type, exc_value, exc_tb)

    return modified_exit


def install():
    """Install the slurm extension to the SimulationWriter class. Extends the
    __exit__ method to add slurm metadata.
    """
    SimulationWriter.__exit__ = _extend_exit_slurm_info(SimulationWriter.__exit__)
    SimulationWriter.__enter__ = _extend_enter_slurm_info(SimulationWriter.__enter__)



import argparse

import pandas as pd
import rich

from bamboost import _config, index
from bamboost.index import DatabaseTable, IndexAPI
from bamboost.manager import Manager
from bamboost.simulation import Simulation
from bamboost.simulation_writer import SimulationWriter


def main():
    parser = argparse.ArgumentParser(description="CLI for bamboost")
    parser.add_argument(
        "--remote", "-r", help="Remote server to fetch data from", type=str
    )
    subparsers = parser.add_subparsers(
        dest="command",
    )
    function_map = {}

    # ----------------
    # Submit
    # ----------------
    function_map["submit"] = submit_simulation
    parser_submit = subparsers.add_parser("submit", help="Submit a job")
    parser_submit.add_argument(
        "path",
        nargs="?",
        type=str,
        help="Path to simulation (directory) to submit",
    )
    parser_submit.add_argument(
        "--id",
        default=None,
        type=str,
        help="Simulation ID",
    )
    parser_submit.add_argument(
        "--db",
        default=None,
        type=str,
        help="Database ID in which a submission will be made",
    )
    parser_submit.add_argument(
        "--all",
        action="store_true",
        help="Submit all unsubmitted simulations in the database",
    )

    # ----------------
    # Database
    # ----------------
    function_map["db"] = manage_db
    parser_db = subparsers.add_parser("db", help="Database management")
    parser_db.add_argument("db_id", type=str, help="Database ID")
    sub_parser_db = parser_db.add_subparsers(
        dest="subcommand",
        help="Database management subcommand",
    )
    sub_parser_db.add_parser("list", help="List all simulations in the database")
    sub_parser_db.add_parser("drop", help="Drop database from the index")
    sub_parser_db.add_parser("reset", help="Reset its table in the sqlite database")

    # ----------------
    # List
    # ----------------
    function_map["list"] = list_databases
    parser_list = subparsers.add_parser("list", help="List all databases")  # noqa: F841

    # ----------------
    # Clean
    # ----------------
    function_map["clean"] = clean_index
    parser_clean = subparsers.add_parser("clean", help="Clean the index")
    parser_clean.add_argument(
        "--purge",
        action="store_true",
        help="also remove the tables from unknown databases",
    )

    # ----------------
    # Scan
    # ----------------
    def scan_known_paths(args):
        IndexAPI().scan_known_paths()
        rich.print("Scanned known paths")

    function_map["scan"] = scan_known_paths

    parser_scan = subparsers.add_parser("scan", help="Scan known paths")  # noqa: F841

    # ----------------
    # Open config file
    # ----------------
    function_map["config"] = open_config
    parser_config = subparsers.add_parser("config", help="Open the config file")
    parser_config.add_argument(
        "--tui", "-t", action="store_true", help="Open the tui config file"
    )
    parser_config.add_argument(
        "--functions", "-f", action="store_true", help="Open the tui functions file"
    )

    # ----------------
    # New database
    # ----------------
    function_map["new"] = create_new_database
    parser_new = subparsers.add_parser("new", help="Create a new database")
    parser_new.add_argument("path", type=str, help="Path to the database")

    # ----------------
    # Parse
    # ----------------
    args = parser.parse_args()
    if args.command is None:
        parser.print_help()
        return
    func = function_map[args.command]
    func(args)


def submit_simulation(args):
    if args.path is not None:
        db_path, uid = args.path.rstrip("/").rsplit("/", 1)
        sim = Simulation(uid, db_path)
        args.id = uid
        args.db = IndexAPI().get_id(db_path)
    else:
        assert args.id and args.db, "Simulation ID and database ID must be provided"
        sim = Simulation(args.id, args.db)

    sim.submit()
    print(f"Submitted simulation {args.id} [db: {args.db}]")


def manage_db(args):
    # test if the database exists
    df = IndexAPI().read_table()
    if args.db_id not in df["id"].values:
        match = df[df["path"].astype(str).str.contains(args.db_id, na=False)].reset_index(drop=True)
        choice = 0
        if len(match) > 1:
            rich.print(
                f"Database ID {args.db_id} is ambiguous. Possible matches: \n{match.to_string()}"
            )
            choice = int(input("Choose one of the above databases to continue: "))
        args.db_id = match.iloc[choice]["id"]

    if args.subcommand == "list":
        rich.print(Manager(uid=args.db_id).df)
    if args.subcommand == "reset":
        with IndexAPI().open():
            IndexAPI()._conn.execute(f"DROP TABLE IF EXISTS db_{args.db_id}")
            IndexAPI()._conn.execute(f"DROP TABLE IF EXISTS db_{args.db_id}_t")
        rich.print(f"Database {args.db_id} dropped")

    if args.subcommand == "drop":
        with IndexAPI().open():
            IndexAPI()._conn.execute(f"DROP TABLE IF EXISTS db_{args.db_id}")
            IndexAPI()._conn.execute(f"DROP TABLE IF EXISTS db_{args.db_id}_t")
        IndexAPI().drop_path(args.db_id)
        rich.print(f"Database {args.db_id} dropped")


def list_databases(args):
    table = IndexAPI().read_table()
    rich.print(table.to_string())


def clean_index(args, purge: bool = False):
    IndexAPI().clean(purge=purge)
    rich.print("Index cleaned")


def open_config(args):
    import os
    import subprocess

    if args.tui:
        subprocess.run(
            [
                f"{os.environ.get('EDITOR', 'vi')}",
                f"{_config.paths['CONFIG_DIR']}/tui.toml",
            ]
        )
    elif args.functions:
        subprocess.run(
            [
                f"{os.environ.get('EDITOR', 'vi')}",
                f"{_config.paths['CONFIG_DIR']}/custom_functions.py",
            ]
        )
    else:
        subprocess.run(
            [f"{os.environ.get('EDITOR', 'vi')}", f"{_config.paths['CONFIG_FILE']}"]
        )


def create_new_database(args):
    try:
        path = index.get_uid_from_path(args.path)
        rich.print(f"Database at {args.path} already exists. ID: {path}")
    except FileNotFoundError:
        db = Manager(args.path, create_if_not_exist=True)
        rich.print(f"Database at {args.path} created. ID: {db.UID}")



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

from __future__ import annotations

from typing import Tuple

import numpy as np

from bamboost import BAMBOOST_LOGGER
from bamboost.common import hdf_pointer
from bamboost.common.file_handler import FileHandler, with_file_open

__all__ = ["MeshGroup", "Mesh"]

log = BAMBOOST_LOGGER.getChild(__name__.split(".")[-1])


class MeshGroup(hdf_pointer.Group):
    def __init__(
        self,
        file_handler: FileHandler,
        path_to_data: str = "/Mesh/0",
        _default_mesh: str = "mesh",
        **kwargs,
    ) -> None:
        super().__init__(file_handler, path_to_data, **kwargs)
        self._default_mesh = _default_mesh

    @with_file_open("r")
    def __getitem__(self, key) -> Mesh:
        return Mesh(self._file, f"{self.path_to_data}/{key}")


class Mesh(hdf_pointer.Group):
    def __init__(self, file_handler: FileHandler, path_to_data: str) -> None:
        super().__init__(file_handler, path_to_data)

    @property
    @with_file_open("r")
    def coordinates(self):
        try:
            return self.obj["geometry"][()]
        except KeyError:
            return self.obj["coordinates"][()]

    @property
    @with_file_open("r")
    def connectivity(self):
        return self.obj["topology"][()]

    @with_file_open("r")
    def get_tuple(self) -> Tuple[np.ndarray, np.ndarray]:
        return (self.coordinates, self.connectivity)



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

from __future__ import annotations

from typing import Generator, Tuple

import h5py
import numpy as np
import pandas as pd

from bamboost import BAMBOOST_LOGGER
from bamboost.accessors.meshes import Mesh, MeshGroup
from bamboost.common import hdf_pointer
from bamboost.common.file_handler import FileHandler, with_file_open

__all__ = ["DataGroup", "FieldData"]

log = BAMBOOST_LOGGER.getChild(__name__.split(".")[-1])


# -----------------------------------------------------------------------------


class DataGroup(hdf_pointer.Group):
    """This pointer points to the data directory. Item accessor returns the
    individual data fields. `meshes` is passed to here for access of linked
    meshes.
    """

    def __init__(
        self,
        file_handler: FileHandler,
        meshes: MeshGroup,
        path_to_data: str = "/data",
        **kwargs,
    ) -> None:
        super().__init__(file_handler, path_to_data, **kwargs)
        self.meshes = meshes

    def __getitem__(self, key) -> FieldData:
        return FieldData(self._file, f"{self.path_to_data}/{key}", meshes=self.meshes)

    def __iter__(self) -> Generator[FieldData, None, None]:
        for key in self.keys():
            yield self.__getitem__(key)

    @property
    @with_file_open("r")
    def info(self) -> pd.DataFrame:
        """View the data stored.

        Returns:
            :class:`pd.DataFrame`
        """
        tmp_dictionary = dict()
        for data in self:
            steps = len(data)
            shape = data.obj["0"].shape
            dtype = data.obj["0"].dtype
            tmp_dictionary[data._name] = {
                "dtype": dtype,
                "shape": shape,
                "steps": steps,
            }
        return pd.DataFrame.from_dict(tmp_dictionary)


class FieldData(hdf_pointer.Group):
    """This pointer points to a specific data field. `meshes` is passed to here
    for access of linked meshes.
    """

    _vds_key = "__vds"  # We create a virtual dataset for slicing across all steps
    _times_key = (
        "__times"  # We store the time of steps in a seperate dataset for performance
    )

    def __init__(
        self, file_handler: FileHandler, path_to_data: str, meshes: MeshGroup
    ) -> None:
        super().__init__(file_handler, path_to_data)
        self.meshes = meshes
        self._name = path_to_data.split("/")[-1]

    @with_file_open("r")
    def __getitem__(self, key) -> np.ndarray:
        return self._get_full_data()[key]

    def _get_full_data(self) -> h5py.Dataset:
        """Return a HDF5 virtual dataset including all steps of the field.

        If the virtual dataset exists with the correct shape it will be
        returned, otherwise it will be created.
        """
        if self._vds_key not in self.keys():
            self._create_vds()
            return self.obj[self._vds_key]

        if len(self) > self.obj.attrs.get("virtual_dataset_length", 0):
            self._create_vds()
            self._create_times()
            return self.obj[self._vds_key]

        return self.obj[self._vds_key]

    @with_file_open("r")
    def __len__(self) -> int:
        return len(self.datasets())

    @with_file_open("r")
    def datasets(self) -> set:
        return super().datasets() - {self._vds_key, self._times_key}

    @property
    @with_file_open("r")
    def shape(self) -> tuple:
        return self._get_full_data().shape

    @property
    @with_file_open("r")
    def dtype(self) -> type:
        return self._get_full_data().dtype

    @with_file_open()
    def at_step(self, *steps: int) -> np.ndarray:
        """Direct access to data at step. Does not require the virtual dataset.

        Args:
            step0, step1, ...: step to extract (can be multiple)
        Returns:
            :class:`np.ndarray`
        """
        data = list()
        for step in steps:
            if step < 0:
                step = len(self) + step
            data.append(self.obj[str(step)][()])
        if len(data) <= 1:
            return data[0]
        else:
            return data

    @property
    @with_file_open()
    def times(self) -> np.ndarray:
        """Return the array of timestamps.

        Returns:
            :class:`np.ndarray`
        """
        if (self._times_key not in self.keys()) or (
            len(self) > self.obj.attrs.get("virtual_dataset_length", 0)
        ):
            self._create_times()

        return self.obj[self._times_key][()]

    @property
    @with_file_open()
    def mesh(self) -> Mesh:
        """Return the linked mesh. Currently returns the linked mesh of first step only.

        Returns:
            :class:`tuple[np.ndarray, np.ndarray]`
        """
        mesh_name = self.obj["0"].attrs.get("mesh", self.meshes._default_mesh)
        return self.meshes[mesh_name]

    @property
    def coordinates(self) -> np.ndarray:
        """Wrapper for mesh.coordinates"""
        return self.mesh.coordinates

    @property
    def connectivity(self) -> np.ndarray:
        """Wrapper for mesh.connectivity"""
        return self.mesh.connectivity

    @property
    @with_file_open()
    def msh(self) -> Tuple[np.ndarray, np.ndarray]:
        """Wrapper to get mesh as tuple"""
        return (self.mesh.coordinates, self.mesh.connectivity)

    def regenerate_virtual_datasets(self) -> None:
        """Regenerate virtual dataset. Call this if the data has changed, thus the
        virtual datasets need to be updated to cover the actual data.
        """
        self._create_times()
        self._create_vds()

    @with_file_open("r+")
    def _create_times(self) -> None:
        times = [self.obj[str(step)].attrs.get("t", step) for step in range(len(self))]
        if self._times_key in self.obj.keys():
            del self.obj[self._times_key]
        self.obj.create_dataset(self._times_key, data=np.array(times))

    def _create_vds(self) -> None:
        """Create virtual dataset of the full timeseries of a field.
        Requires HDF5 > 1.10 !
        """
        with self._file("r"):
            length = len(self)
            ds_shape = self.obj["0"].shape
            ds_dtype = self.obj["0"].dtype

            layout = h5py.VirtualLayout(shape=(length, *ds_shape), dtype=ds_dtype)

            for step in range(length):
                vsource = h5py.VirtualSource(self.obj[str(step)])
                layout[step] = vsource

        with self._file("r+"):
            if self._vds_key in self.obj.keys():
                del self.obj[self._vds_key]
            self.obj.attrs["virtual_dataset_length"] = length
            self.obj.create_virtual_dataset(self._vds_key, layout)



import os
import shutil
import tempfile

os.environ["BAMBOOST_MPI"] = "0"
from bamboost import index
from bamboost._config import paths


def pytest_sessionstart(session):
    """Setup tmp config directory."""
    tempdir = tempfile.mkdtemp()
    paths["CONFIG_DIR"] = tempdir
    paths["CONFIG_FILE"] = os.path.join(tempdir, "config.toml")
    paths["LOCAL_DIR"] = os.path.join(tempdir, "local")
    paths["DATABASE_FILE"] = os.path.join(tempdir, "local", "bamboost.db")

    # Create config files if they don't exist
    os.makedirs(paths["CONFIG_DIR"], exist_ok=True)
    os.makedirs(paths["LOCAL_DIR"], exist_ok=True)

    # change path of sqlite database
    # index.IndexAPI(_file=).__init__()
    print(index.IndexAPI().file)


def pytest_sessionfinish(session, exitstatus):
    """Remove tmp config directory again."""
    shutil.rmtree(paths["CONFIG_DIR"])



import os

import numpy as np
import pytest
from test_manager import temp_manager


def test_enter_path(temp_manager):
    sim = temp_manager.create_simulation()
    sim.finish_sim()
    path = sim.path
    uid = sim.uid
    original_path = os.getcwd()
    with temp_manager.sim(uid).enter_path():
        assert os.getcwd() == path
    assert os.getcwd() == original_path

    # check that we return in the right path even when throwing an error
    try:
        with temp_manager.sim(uid).enter_path():
            raise ValueError()
    except ValueError:
        assert os.getcwd() == original_path


@pytest.mark.parametrize(
    "data",
    [
        np.array([1, 2, 3]),
        np.array([1.2, 2.3, 3.4]),
        np.array([[1, 2], [3, 4]]),
    ],
)
def test_read_field(temp_manager, data):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.add_field("test", data)
    sim.finish_step()

    sim = temp_manager.sim(uid)
    data_read = sim.data["test"].at_step(0)
    assert np.array_equal(data_read, data)



"""
The module includes functions to copy an example config file to the user's
config directory and load the config file from the directory. If the config
file does not exist, a copy of the example config file will be created.

Functions:
    - _copy_config_file(): Copies the example config file to the user's config
      directory.
    - _load_config_file(): Loads the config file from the user's config
      directory.

Attributes:
    - paths: A dictionary containing paths to the config files.
    - config: A dictionary of the config. Initiated from config file.
"""

import os
import pkgutil

try:
    import tomllib as tomli
except ImportError:
    import tomli

__all__ = ["paths", "config"]

# Define paths to bamboost config files
_home_dir = os.path.expanduser("~")
paths = {
    "HOME": os.path.expanduser("~"),
    "CONFIG_DIR": os.path.join(_home_dir, ".config", "bamboost"),
    "LOCAL_DIR": os.path.join(_home_dir, ".local", "share", "bamboost"),
    "CONFIG_FILE": os.path.join(_home_dir, ".config", "bamboost", "config.toml"),
    "DATABASE_FILE": os.path.join(
        _home_dir, ".local", "share", "bamboost", "bamboost.db"
    ),
}

# Create directories if they do not exist
os.makedirs(paths["CONFIG_DIR"], exist_ok=True)
os.makedirs(paths["LOCAL_DIR"], exist_ok=True)


def _copy_config_file(file: str):
    """Copy the example config file to the user's config directory.

    Raises:
        - FileExistsError: If the config file already exists in the destination
          directory."""
    # If file exists, raise an error
    if os.path.exists(file):
        raise FileExistsError("Config file already exists")

    # source_file = "_example_config.toml"
    source_file: bytes = pkgutil.get_data("bamboost", "_example_config.toml")

    # Copy the source file to the destination
    with open(file, "wb") as f:
        f.write(source_file)


def _load_config_file(file: str = None):
    """Load the config file from the user's config directory.

    If the config file does not exist, a copy of the example config file will
    be created.

    Returns:
        - dict: The contents of the config file as a dictionary.
    """
    # Load the config file
    if file is None:
        file = paths["CONFIG_FILE"]
    try:
        with open(file, "rb") as f:
            return tomli.load(f)
    except FileNotFoundError:
        _copy_config_file(file)
        return _load_config_file(file)


config = _load_config_file()



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code
"""Module to manage the database index and its ID's.

Attributes:
    THREAD_SAFE: if True, the index is thread safe
    CONVERT_ARRAYS: if True, convert numpy arrays to lists
"""

from __future__ import annotations

from bamboost import BAMBOOST_LOGGER

__all__ = [
    "Null",
    "IndexAPI",
    "DatabaseTable",
    "Entry",
    "find",
    "get_uid_from_path",
    "get_known_paths",
    "uid2",
    "DatabaseNotFoundError",
    # Constants
    "THREAD_SAFE",
    "CONVERT_ARRAYS",
    "PREFIX",
    "DOT_REPLACEMENT",
]

import json
import os
import subprocess
from dataclasses import dataclass
from time import time
from typing import Callable
from functools import wraps

import pandas as pd

import bamboost._sqlite_database as sql
from bamboost._config import config, paths
from bamboost.common.file_handler import open_h5file
from bamboost.common.mpi import MPI

log = BAMBOOST_LOGGER.getChild(__name__.split(".")[-1])


PREFIX = ".BAMBOOST-"
"prefix for databaseID identifier file"

DOT_REPLACEMENT = "DOT"  
"replace dots with this in column names for sqlite"

_comm = MPI.COMM_WORLD  # TODO: is this good practice?

THREAD_SAFE = False
CONVERT_ARRAYS = True


# ------------------
# Exceptions
# ------------------
class DatabaseNotFoundError(Exception):
    """Exception raised when a database is not found in the index."""

    pass


# ------------------
# Classes
# ------------------

Error = sql.sqlite3.Error
"""Error exception for index errors."""


class Null:
    """Null object to replace API classes for off-root processes."""

    def __getattr__(self, _):
        return self

    def __bool__(self):
        # Allows the instance to behave like `None` in boolean contexts
        return False

    def __call__(self, *args, **kwargs):
        # Allows the instance to be called like a function
        return self

    def __getitem__(self, _):
        return self

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        return False


class IndexAPI(sql.SQLiteHandler):
    """SQLite database to store database ID, path lookup. As well as the table for
    each database. Location: `~/.local/share/bamboost/bamboost.db`.
    Singleton pattern.

    Args:
        _file: path to the database file
        convert_arrays: Defaults to True. If False, when reading from the database,
            lists with tag ARRAY are not converted back to numpy arrays but
            remain a standard list.
    """

    _instances = {}

    def __new__(cls, *, _file: str = None, **kwargs) -> IndexAPI:
        if THREAD_SAFE:
            return super().__new__(cls)

        if _comm.rank != 0:
            return Null()

        _file = _file or paths["DATABASE_FILE"]
        if _file not in cls._instances:
            cls._instances[_file] = super().__new__(cls)
        return cls._instances[_file]

    def __init__(self, *, _file: str = None, convert_arrays: bool = None):
        convert_arrays = (
            convert_arrays if convert_arrays is not None else CONVERT_ARRAYS
        )
        if hasattr(self, "_initialized"):
            return
        _file = _file or paths["DATABASE_FILE"]
        super().__init__(file=_file, convert_arrays=convert_arrays)
        self.create_index_table()
        self.clean()
        self._initialized = True

    @classmethod
    def ThreadSafe(cls, *args, **kwargs) -> IndexAPI:
        instance = super().__new__(cls)
        instance.__init__(*args, **kwargs)
        return instance

    def __repr__(self) -> str:
        return self.read_table().__repr__()

    def _repr_html_(self) -> str:
        return self.read_table()._repr_html_()

    def __getitem__(self, id: str) -> DatabaseTable:
        return DatabaseTable(id, _index=self)

    def _ipython_key_completions_(self) -> list:
        return self.read_table().id.tolist()

    @sql.with_connection
    def create_index_table(self) -> None:
        """Create the index table if it does not exist."""
        self._cursor.execute(
            """CREATE TABLE IF NOT EXISTS dbindex (id TEXT PRIMARY KEY, path TEXT)"""
        )

    def get_database_table(self, id: str) -> DatabaseTable:
        """Get the table of a database.

        Args:
            - id (str): ID of the database
        """
        return DatabaseTable(id, _index=self)

    @sql.with_connection
    def read_table(self, *args, **kwargs) -> pd.DataFrame:
        """Read the index table.

        Returns:
            pd.DataFrame: index table
        """
        return pd.read_sql_query("SELECT * FROM dbindex", self._conn, *args, **kwargs)

    @sql.with_connection
    def fetch(self, query: str, *args, **kwargs) -> pd.DataFrame:
        """Query the index table.

        Args:
            query (str): query string
        """
        self._cursor.execute(query, *args, **kwargs)
        return self._cursor.fetchall()

    @sql.with_connection
    def get_path(self, id: str) -> str:
        """Get the path of a database from its ID.

        Args:
            id (str): ID of the database

        Returns:
            str: path of the database
        """
        self._cursor.execute("SELECT path FROM dbindex WHERE id=?", (id,))
        path_db = self._cursor.fetchone()
        if path_db and _check_path(id, path_db[0]):
            return path_db[0]

        # if path is wrong, try to find it
        for root_dir in get_known_paths():
            res = find(id, root_dir)
            if res:
                path = os.path.dirname(res[0])
                self.insert_path(id, path)
                return path

        # last resort, check home
        res = find(id, paths["HOME"])
        if res:
            path = os.path.dirname(res[0])
            self.insert_path(id, path)
            return path

        raise FileNotFoundError(f"Database {id} not found on system.")

    @sql.with_connection
    def insert_path(self, id: str, path: str) -> None:
        """Insert a database path into the index.

        Args:
            id (str): ID of the database
            path (str): path of the database
        """
        path = os.path.abspath(path)
        self._cursor.execute("INSERT OR REPLACE INTO dbindex VALUES (?, ?)", (id, path))

    @sql.with_connection
    def get_id(self, path: str) -> str:
        """Get the ID of a database from its path.

        Args:
            path (str): path of the database

        Returns:
            str: ID of the database

        Raises:
            DatabaseNotFoundError: if the database is not found in the index
        """
        path = os.path.abspath(path)
        self._cursor.execute("SELECT id FROM dbindex WHERE path=?", (path,))
        fetched = self._cursor.fetchone()
        if fetched is None:
            raise DatabaseNotFoundError(f"Database at {path} not found in index.")
        return fetched[0]

    @sql.with_connection
    def scan_known_paths(self) -> dict:
        """Scan known paths for databases and update the index."""
        for path in get_known_paths():
            completed_process = subprocess.run(
                ["find", path, "-iname", f"{PREFIX}*", "-not", "-path", r"*/\.git/*"],
                capture_output=True,
            )
            databases_found = completed_process.stdout.decode("utf-8").splitlines()
            for database in databases_found:
                name = os.path.basename(database)
                id = name.split("-")[1]
                self.insert_path(id, os.path.dirname(database))

    def commit_once(self, func) -> Callable:
        """Decorator to bundle changes to a single commit.

        Example:
            >>> @Index.commit_once
            >>> def create_a_bunch_of_simulations():
            >>>     for i in range(1000):
            >>>         db.create_simulation(parameters={...})
            >>>
            >>> create_a_bunch_of_simulations()
        """

        def wrapper(*args, **kwargs):
            with self.open(ensure_commit=True):
                return func(*args, **kwargs)

        return wrapper

    @sql.with_connection
    def clean(self, purge: bool = False) -> IndexAPI:
        """Clean the index from wrong paths.

        Args:
            purge (bool, optional): Also deletes the table of unmatching uid/path pairs. Defaults to False.
        """
        index = self._cursor.execute("SELECT id, path FROM dbindex").fetchall()
        for id, path in index:
            if not _check_path(id, path):
                self._cursor.execute("DELETE FROM dbindex WHERE id=?", (id,))

        if purge:
            # all tables starting with db_ are tables of databases
            all_tables = self._cursor.execute(
                "SELECT name FROM sqlite_master WHERE type='table'"
            ).fetchall()
            id_list_tables = {
                i[0].split("_")[1] for i in all_tables if i[0].startswith("db_")
            }
            id_list = self._cursor.execute("SELECT id FROM dbindex").fetchall()

            for id in id_list_tables:
                if id not in id_list:
                    self._cursor.execute(f"DROP TABLE db_{id}")
                    self._cursor.execute(f"DROP TABLE db_{id}_t")

        return self

    @sql.with_connection
    def drop_path(self, id: str) -> None:
        """Drop a path from the index.

        Args:
            id (str): ID of the database
        """
        self._cursor.execute("DELETE FROM dbindex WHERE id=?", (id,))

    def check_path(self, id: str, path: str) -> bool:
        """Check if path is going to the correct database."""
        return _check_path(id, path)


class DatabaseTable:
    """
    Class to manage the table of a database. Multiton pattern. One table per
    database.
    """

    _instances = {}

    def __new__(cls, id: str, *args, **kwargs) -> DatabaseTable:
        if _comm.rank != 0:
            return Null()

        if id not in cls._instances:
            cls._instances[id] = super().__new__(cls)
        return cls._instances[id]

    def __init__(self, id: str, *, _index: IndexAPI = None):
        if hasattr(self, "_initialized"):
            return

        self.id = id
        self._entries = {}
        self._initialized = True
        self._index = _index if _index is not None else IndexAPI()
        self.path = self._index.get_path(self.id)
        self.tablename_db = f"db_{self.id}"
        self.tablename_update_times = f"db_{self.id}_t"
        self.create_database_table()

    def __getattr__(self, name):
        if name in {
            "_conn",
            "_cursor",
            "open",
            "close",
            "commit",
            "_is_open",
            "commit_once",
        }:
            return getattr(self._index, name)
        return self.__getattribute__(name)

    # ---------------------
    # Database table functions
    # ---------------------
    @sql.with_connection
    def read_table(self) -> pd.DataFrame:
        """Read the table of the database.

        Returns:
            pd.DataFrame: table of the database
        """
        df = pd.read_sql_query(f"SELECT * FROM {self.tablename_db}", self._conn)
        # drop "hidden" columns which start with _
        df = df.loc[:, ~df.columns.str.startswith("_")]
        df.rename(columns=lambda x: x.replace(DOT_REPLACEMENT, "."), inplace=True)
        return df

    @sql.with_connection
    def read_entry(self, entry_id: str) -> pd.Series:
        """Read an entry from the database.

        Args:
            entry_id (str): ID of the entry

        Returns:
            pd.Series: entry from the database
        """
        self._cursor.execute(
            f"SELECT * FROM {self.tablename_db} WHERE id=?", (entry_id,)
        )
        series = pd.Series(*self._cursor.fetchall())
        series.index = [description[0] for description in self._cursor.description]
        series.rename(index=lambda x: x.replace(DOT_REPLACEMENT, "."), inplace=True)
        return series

    @sql.with_connection
    def read_column(self, *columns: str) -> pd.DataFrame:
        """Read columns from the database.

        Args:
            *columns (list): columns to read

        Returns:
            pd.DataFrame: columns from the database
        """
        self._cursor.execute(f"SELECT {', '.join(columns)} FROM {self.tablename_db}")
        df = pd.DataFrame.from_records(self._cursor.fetchall(), columns=columns)
        df.rename(columns=lambda x: x.replace(DOT_REPLACEMENT, "."), inplace=True)
        return df

    @sql.with_connection
    def drop_table(self) -> None:
        """Drop the table of the database."""
        self._cursor.execute(f"DROP TABLE {self.tablename_db}")
        self._cursor.execute(f"DROP TABLE {self.tablename_update_times}")

    @sql.with_connection
    def create_database_table(self) -> None:
        """Create a table for a database."""
        self._cursor.execute(
            f"""CREATE TABLE IF NOT EXISTS {self.tablename_db} 
                (id TEXT PRIMARY KEY NOT NULL, time_stamp DATETIME, notes TEXT, processors INTEGER)
            """
        )
        self._cursor.execute(
            f"""CREATE TABLE IF NOT EXISTS {self.tablename_update_times} (id TEXT PRIMARY KEY,
                update_time DATETIME)
            """
        )

    @sql.with_connection
    def update_entry(self, entry_id: str, data: dict) -> None:
        """Update an entry in the database.

        Args:
            entry_id (str): ID of the entry
            data (dict): data to update
        """
        # return if data is empty
        if not data:
            return

        # get columns of table
        self._cursor.execute(f"PRAGMA table_info({self.tablename_db})")
        cols = self._cursor.fetchall()

        # replace dots in keys
        for key in list(data.keys()):
            new_key = key.replace(".", DOT_REPLACEMENT)
            new_key = _remove_illegal_column_characters(new_key)
            if new_key != key:
                data[new_key] = data.pop(key)

        # check if columns exist
        for key, val in data.items():
            # key = key.replace(".", DOT_REPLACEMENT)
            # key = _remove_illegal_column_characters(key)
            if any(key == column[1] for column in cols):
                continue
            dtype = sql.get_sqlite_column_type(val)
            self._cursor.execute(
                f"ALTER TABLE {self.tablename_db} ADD COLUMN [{key}] {dtype}"
            )

        # insert data into table
        data.pop("id", None)

        keys = ", ".join([f"[{key}]" for key in data.keys()])
        values = ", ".join([f":{key}" for key in data.keys()])
        updates = ", ".join([f"[{key}] = excluded.[{key}]" for key in data.keys()])

        query = f"""
        INSERT INTO {self.tablename_db} (id, {keys})
        VALUES (:id, {values})
        ON CONFLICT(id) DO UPDATE SET
        {updates}
        """
        data["id"] = entry_id
        self._cursor.execute(query, data)

        # update update time
        self._cursor.execute(
            f"INSERT OR REPLACE INTO {self.tablename_update_times} VALUES (?, ?)",
            (entry_id, time()),
        )

    @sql.with_connection
    def sync(self) -> None:
        """Sync the table with the file system."""
        all_ids_fs = set(
            [
                i
                for i in os.listdir(self.path)
                if os.path.isdir(os.path.join(self.path, i))
            ]
        )

        self._cursor.execute(
            f"SELECT id, update_time FROM {self.tablename_update_times}"
        )

        for id, last_up_time in self._cursor.fetchall():
            # remove entries that do not exist on the file system
            if id not in all_ids_fs:
                self._cursor.execute(
                    f"DELETE FROM {self.tablename_db} WHERE id=?", (id,)
                )
                self._cursor.execute(
                    f"DELETE FROM {self.tablename_update_times} WHERE id=?", (id,)
                )
                continue

            # update entries that have been modified
            all_ids_fs.remove(id)
            if self.entry(id).mtime > last_up_time:
                self.update_entry(id, self.entry(id).get_all_metadata())

        # add new entries
        for id in all_ids_fs:
            self.update_entry(id, self.entry(id).get_all_metadata())

    @sql.with_connection
    def entry(self, entry_id: str) -> Entry:
        """Get the Entry object of an entry.
        Multiton pattern. One Entry per entry.
        """
        if entry_id not in self._entries:
            self._entries[entry_id] = Entry(entry_id, self.path)
        return self._entries[entry_id]


def capture_fileIO_error(method):
    @wraps(method)
    def inner(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except OSError as e:
            e.add_note(f"[h5file: {self.h5file}, ]")
            raise e
        

    return inner

@dataclass
class Entry:
    """Simulation entry in a database.
    Simplified version of the Simulation class in the simulation module.
    """

    def __init__(self, id: str, path: str) -> None:
        self.id = id
        self.path = path
        self.h5file = os.path.join(self.path, self.id, f"{self.id}.h5")

    @property
    @capture_fileIO_error
    def metadata(self) -> dict:
        """Get the metadata of the entry."""
        with open_h5file(self.h5file, "r") as file:
            return dict(file.attrs)

    @property
    @capture_fileIO_error
    def parameters(self) -> dict:
        """Get the parameters of the entry."""
        tmp_dict = dict()
        with open_h5file(self.h5file, "r") as file:
            try:
                tmp_dict.update(file["parameters"].attrs)
                for key in file["parameters"].keys():
                    tmp_dict.update({key: file[f"parameters/{key}"][()]})
            except KeyError:
                pass

        return tmp_dict

    def get_all_metadata(self) -> dict:
        """Get all metadata of the entry."""
        return {**self.metadata, **self.parameters}

    @property
    def mtime(self) -> float:
        """Get the modification time of the entry."""
        return os.path.getmtime(self.h5file)


# ----------------------
# Module level functions
# ----------------------


def find(uid, root_dir) -> list:
    """Find the database with UID under given root_dir.

    Args:
        uid: UID to search for
        root_dir: root directory for search
    """
    if os.name == "posix":
        paths = _find_posix(uid, root_dir)
    else:
        paths = _find_python(uid, root_dir)
    if len(paths) > 1:
        log.warning(f"Multiple paths found for UID {uid}:\n{paths}")
    return paths


def get_uid_from_path(path: str) -> str:
    """Returns the UID found in the specified path."""
    for file in os.listdir(path):
        if file.startswith(".BAMBOOST"):
            return file.split("-")[1]
    raise FileNotFoundError("No UID file found at specified path.")


def get_known_paths() -> list:
    return config["index"].get("paths", [])


def _find_posix(uid, root_dir) -> list:
    """Find function using system `find` on linux."""
    completed_process = subprocess.run(
        ["find", root_dir, "-iname", uid2(uid), "-not", "-path", r"*/\.git/*"],
        capture_output=True,
    )
    paths_found = completed_process.stdout.decode("utf-8").splitlines()
    return paths_found


def _find_python(uid, root_dir) -> list:
    """Some find function for Windows or other if `find` is not working.

    TODO: to be implemented
    """
    pass


def uid2(uid) -> str:
    return f"{PREFIX}{uid}"


def _check_path(uid: str, path: str) -> bool:
    """Check if path is going to the correct database"""
    if not os.path.exists(path):
        return False
    if f"{PREFIX}{uid}" in os.listdir(path):
        return True
    return False


def _remove_illegal_column_characters(key: str) -> str:
    """Remove illegal characters in sqlite column names from a string.

    Args:
        key (str): key to clean

    Removes all content in parenthesis and replaces dashes with underscores.
    """
    import re

    # clean from parenthesis
    key = re.sub(r"[()]", "", re.sub(r"\(.*?\)", "", key))
    # clean from dashes
    key = re.sub(r"-", "_", key)
    return key



"""This module provides a custom rather primitive file locking mechanism for
HDF5 files.

This module provides a function `use_locking` that installs the mechanism by
Monkey-patching the `FileHandler` class. To use this mechanism, call the
`use_locking` function at the start of your script, before creating any
`bamboost.common.file_handler.FileHandler` instances.

Example:
    >>> from bamboost.extensions.use_locking import use_locking
    >>> use_locking("shared")
"""

from __future__ import annotations

import fcntl
import os
from collections.abc import Callable
from functools import wraps
from io import BufferedReader
from typing import Any, Literal, Tuple

import h5py

from bamboost.common.file_handler import (
    FILE_MODE_HIRARCHY,
    FileHandler,
    log,
    open_h5file,
)


def get_lock_and_open_function(
    lock_type: Literal["shared", "exclusive"],
) -> Callable[..., Tuple[h5py.File, BufferedReader]]:
    """Returns a function that locks and opens an HDF5 file with the specified
    lock type.

    The returned function acquires a lock on the lock file and opens the HDF5
    file, returning both the opened HDF5 file object and the lock file object.

    The returned function takes the following parameters:

    | Argument  | Description                                                                 |
    |-----------|-----------------------------------------------------------------------------|
    | file      | The path to the HDF5 file to open.                                          |
    | lock_file | The path to the lock file.                                                  |
    | mode      | The mode to open the file in (e.g., 'r', 'w', 'a').                         |
    | driver    | The HDF5 driver to use (e.g., 'mpio' for MPI-IO).                           |
    | comm      | The MPI communicator (only used with 'mpio' driver).                        |

    Args:
        lock_type (Literal["shared", "exclusive"]): The type of lock to use.

    Returns:
        Callable: A function that locks and opens an HDF5 file.
    """
    lock_type_int = {"shared": fcntl.LOCK_SH, "exclusive": fcntl.LOCK_EX}[lock_type]

    def lock_and_open(
        file: str, lock_file: str, mode: str, driver: str, comm: Any
    ) -> Tuple[h5py.File, BufferedReader]:
        lock_file_object = open(lock_file, "rb")
        # acquire a lock
        fcntl.flock(lock_file_object, lock_type_int)
        # open the file
        h5_file = open_h5file(file, mode, driver, comm)
        return h5_file, lock_file_object

    return lock_and_open


def close_and_unlock(h5_file: h5py.File, lock_file: BufferedReader) -> None:
    """Close an HDF5 file and release the associated file lock.

    This function performs the following operations:
    1. Closes the given HDF5 file.
    2. Releases the lock on the lock file.
    3. Closes the lock file.

    Args:
        h5_file: The HDF5 file object to be closed.
        lock_file: The lock file object to be unlocked and closed.

    Returns:
        None
    """
    h5_file.close()
    fcntl.flock(lock_file, fcntl.LOCK_UN)
    lock_file.close()


def get_open_method(
    lock_type: Literal["shared", "exclusive"],
) -> Callable[..., h5py.File]:
    open_function = get_lock_and_open_function(lock_type)

    @wraps(FileHandler.open)
    def _modified_open(self: FileHandler, mode: str = "r", driver=None, comm=None):
        """Create a primitive lock file for the file.

        Will replace `file_handler.FileHandler.open` method.
        """
        if self._lock <= 0:
            # lock_file_access(self.file_name, comm)

            log.debug(f"[{id(self)}] Open {self.file_name}")
            # self.file_object = open_h5file(self.file_name, mode, driver, comm)
            self.file_object, self._lock_file = open_function(  # type: ignore
                self.file_name, self._lock_file_name, mode, driver, comm
            )

        if FILE_MODE_HIRARCHY[self.file_object.mode] < FILE_MODE_HIRARCHY[mode]:
            close_and_unlock(self.file_object, self._lock_file)
            self.file_object, self._lock_file = open_function(  # type: ignore
                self.file_name, self._lock_file_name, mode, driver, comm
            )

        log.debug(f"[{id(self)}] Lock stack {self._lock}")
        self._lock += 1
        return self.file_object

    return _modified_open


@wraps(FileHandler.close)
def _modified_close(self: FileHandler):
    """Create a primitive lock for the file.

    Will replace `file_handler.FileHandler.close` method.
    """
    self._lock -= 1
    if self._lock == 0:
        log.debug(f"[{id(self)}] Close {self.file_name}")
        # self.file_object.close()
        close_and_unlock(self.file_object, self._lock_file)

    log.debug(f"[{id(self)}] Lock stack {self._lock}")


def _extend_init(original_init):
    @wraps(FileHandler.__init__)
    def _modified_init(self: FileHandler, *args, **kwargs):
        original_init(self, *args, **kwargs)

        self._lock_file_name = f"{self.file_name}.lock"  # type: ignore

        if not os.path.exists(self._lock_file_name):
            with open(self._lock_file_name, "w") as f:
                f.write("LOCK FILE")

    return _modified_init


def use_locking(lock_type: Literal["shared", "exclusive"]) -> None:
    """Installs a primitive locking mechanism for FileHandler operations.

    This function modifies the FileHandler class to include file locking
    capabilities, ensuring thread-safe access to files. It wraps the
    `__init__`, `open`, and `close` methods of FileHandler with locking mechanisms.

    Usage:
        Call this function before using FileHandler to enable locking:
        use_locking("shared")  # or use_locking("exclusive")

    Note:
        This function should be called only once, preferably at the start
        of your program, before any FileHandler instances are created.

    Note:
        You should only need this functionality if you are using HDF5 <= 1.10.6.
        The newer versions of HDF5 (>= 1.10.7) have built-in file locking.

    Args:
        lock_type (Literal["shared", "exclusive"]): The type of lock to use.
            - "shared" allows multiple processes to open the file simultaneously.
            - "exclusive" allows only one reader or writer at a time.
    """
    if not hasattr(FileHandler.__init__, "__wrapped__"):
        FileHandler.__init__ = _extend_init(FileHandler.__init__)

    if not hasattr(FileHandler.open, "__wrapped__"):
        FileHandler.open = get_open_method(lock_type)

    if not hasattr(FileHandler.close, "__wrapped__"):
        FileHandler.close = _modified_close

    log.info("Primitive locking installed")



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

from __future__ import annotations

import datetime
import os
import shutil
from typing import Any, Dict, Literal, Tuple, Union

import numpy as np
from typing_extensions import deprecated

from bamboost import BAMBOOST_LOGGER
from bamboost.common.git_utility import GitStateGetter
from bamboost.common.mpi import MPI
from bamboost.common.utilities import flatten_dict
from bamboost.simulation import Simulation

__all__ = ["SimulationWriter"]

log = BAMBOOST_LOGGER.getChild(__name__.split(".")[-1])


class SimulationWriter(Simulation):
    """The SimulationWriter is the writer object for a single simulation. It inherits
    all reading methods from :class:`Simulation`.

    This class can be used as a context manager. When entering the context, the status
    of the simulation is changed to "Started". When an exception is raised inside the
    context, the status is changed to "Failed [Exception]".

    Args:
        uid: The identifier of the simulation
        path: The (parent) database path
        comm: An MPI communicator (Default: `MPI.COMM_WORLD`)
    """

    def __init__(
        self,
        uid: str,
        path: str,
        comm: MPI.Comm = MPI.COMM_WORLD,
        create_if_not_exists: bool = True,
    ):
        super().__init__(uid, path, comm, create_if_not_exists)
        self.step: int = 0

    def __enter__(self):
        self.change_status("Started")  # change status to running (process 0 only)
        self._comm.barrier()  # wait for change status to be written
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.change_status(f"Failed [{exc_type.__name__}]")
            log.error(f"Simulation failed with {exc_type.__name__}: {exc_val}")
            log.error(exc_tb)
            raise RuntimeError(f"Simulation failed [Process {self._prank}]")
        self._comm.barrier()

    def initialize(self) -> SimulationWriter:
        """Create a new file for this simlation.
        This deletes an existing h5 file of the simulation and creates an empty new one
        """
        self.step = 0
        self.add_metadata()
        self.change_status("Initiated")

        return self

    def add_metadata(self) -> None:
        """Add metadata to h5 file."""
        nb_proc = self._comm.Get_size()
        if self._prank == 0:
            with self._file("a"):
                data = {
                    "time_stamp": str(datetime.datetime.now().replace(microsecond=0)),
                    "id": self.uid,
                    "processors": nb_proc,
                    "notes": self._file.attrs.get("notes", ""),
                }
                self._file.attrs.update(data)
            self._push_update_to_sqlite(data)

    def add_parameters(self, parameters: dict) -> None:
        """Add parameters to simulation.

        Args:
            parameters: Dictionary with parameters.
        """
        if self._prank == 0:
            with self._file("a"):
                # flatten parameters
                parameters = flatten_dict(parameters)

                if "parameters" in self._file.keys():
                    del self._file["parameters"]
                grp = self._file.create_group("/parameters")
                for key, val in parameters.items():
                    if isinstance(val, np.ndarray):
                        grp.create_dataset(key, data=val)
                    elif val is not None:
                        grp.attrs[key] = val
            self._push_update_to_sqlite(parameters)

    def add_mesh(
        self, coordinates: np.ndarray, connectivity: np.ndarray, mesh_name: str = None
    ) -> None:
        """Add the mesh to file. Currently only 2d meshes.

        Note:
            If you are working with FEniCS, you can use the
            `bamboost.extensions.fenics.FenicsWriter` writer. With it you can
            write the FEniCS mesh directly.
        Args:
            coordinates: Coordinates as array (nb_nodes, dim)
            connectivity: Connectivity matrix (nb_cells, nb nodes per cell)
            mesh_name: name for mesh (default = `mesh`)
        """
        if mesh_name is None:
            mesh_name = self._default_mesh
        # self._mesh_location = 'Mesh/0/mesh/'
        mesh_location = f"{self._mesh_location}/{mesh_name}/"

        nb_nodes_local = coordinates.shape[0]
        nb_cells_local = connectivity.shape[0]

        # gather total mesh
        nb_nodes_p = np.array(self._comm.allgather(nb_nodes_local))
        nb_cells_p = np.array(self._comm.allgather(nb_cells_local))
        nb_nodes, nb_cells = np.sum(nb_nodes_p), np.sum(nb_cells_p)

        # shape of datasets
        coord_shape = (
            (nb_nodes, coordinates.shape[1]) if coordinates.ndim > 1 else (nb_nodes,)
        )
        conn_shape = (
            (nb_cells, connectivity.shape[1]) if connectivity.ndim > 1 else (nb_cells,)
        )

        # global indices nodes
        idx_start = np.sum(nb_nodes_p[self._ranks < self._prank])
        idx_end = idx_start + nb_nodes_local

        # global indices cells
        idx_start_cells = np.sum(nb_cells_p[self._ranks < self._prank])
        idx_end_cells = idx_start_cells + nb_cells_local
        connectivity = connectivity + idx_start

        with self._file("a", driver="mpio", comm=self._comm) as f:
            if mesh_location in self._file.file_object:
                del self._file.file_object[mesh_location]
            grp = f.require_group(mesh_location)
            coord = grp.require_dataset(
                "coordinates", shape=coord_shape, dtype=coordinates.dtype
            )
            conn = grp.require_dataset(
                "topology", shape=conn_shape, dtype=connectivity.dtype
            )

            coord[idx_start:idx_end] = coordinates
            conn[idx_start_cells:idx_end_cells] = connectivity

    def add_field(
        self,
        name: str,
        vector: np.array,
        time: float = None,
        mesh: str = None,
        dtype: str = None,
        center: Literal["Node", "Cell"] = "Node",
    ) -> None:
        """Add a dataset to the file. The data is stored at `data/`.

        Args:
            name: Name for the dataset
            vector: Dataset
            time: Optional. time
            mesh: Optional. Linked mesh for this data
            dtype: Optional. Numpy style datatype, see h5py documentation,
                defaults to the dtype of the vector.
            center: Optional. Center of the data. Can be 'Node' or 'Cell'.
                Default is 'Node'.
        """
        if mesh is None:
            mesh = self._default_mesh

        if time is None:
            time = self.step

        self._dump_array(f"data/{name}/{self.step}", vector, dtype=dtype)
        self._comm.barrier()  # attempt to fix bug (see SimulationWriter add_field)

        if self._prank == 0:
            with self._file("a"):
                # Sometimes this fails with (if simultaneously trying to read the file)
                # KeyError: 'Unable to synchronously open object (addr overflow, addr = 247903512, size = 328, eoa = 247903240)'
                # KeyError: 'Unable to synchronously open object (len not positive after adjustment for EOA)'
                # I don't know exactly what is going on.
                # It could be that some process is still writing and then it's opened again and the dataset doesn't exist properly
                # OR the other processes try to open the file already for the next time while this one is waiting
                vec = self._file["data"][name][str(self.step)]
                vec.attrs.update({"center": center, "mesh": mesh, "t": time})

        self._comm.barrier()  # attempt to fix bug (see SimulationWriter add_field)

    def _dump_array(self, location: str, arr: np.ndarray, dtype: str = None) -> None:
        """Dump an array to the file. Correctly patch together multi-rank arrays.

        Args:
            location: Location in the file
            arr: Array to dump
            dtype: Optional. Numpy style datatype, see h5py documentation,
                defaults to the dtype of the vector.
        """
        dim = arr.shape[1:] if arr.ndim > 1 else None
        length_local = arr.shape[0]
        length_p = np.array(self._comm.allgather(length_local))
        length = np.sum(length_p)

        # split location into group and dataset
        group_name, dataset_name = location.rstrip("/").rsplit("/", 1)

        # global indices
        idx_start = np.sum(length_p[self._ranks < self._prank])
        idx_end = idx_start + length_local

        # open file
        with self._file("a", driver="mpio", comm=self._comm) as f:
            grp = f.require_group(group_name)
            vec = grp.require_dataset(
                dataset_name,
                shape=(length, *dim) if dim else (length,),
                dtype=dtype if dtype else arr.dtype,
            )
            vec[idx_start:idx_end] = arr

    def add_fields(
        self,
        fields: Dict[str, np.ndarray | Tuple[np.ndarray, str]],
        time: float = None,
        mesh: str = None,
    ) -> None:
        """Add multiple fields at once.

        Args:
            fields: Dictionary with fields. The value can be a tuple with the
                data and a string "Node" or "Cell".
            time: Optional. time
        """
        for key, value in fields.items():
            if isinstance(value, tuple):
                vector, center = value
            else:
                vector, center = value, "Node"
            self.add_field(key, vector, time, mesh, center=center)

    def add_global_field(self, name: str, value: Any, dtype: str = None) -> None:
        """Add a gobal field. These are stored at `globals/` as an array in a
        single dataset.

        Args:
            name: Name for the data
            value: Data. Can be a numpy array or a single value.
        """
        self._dump_global_data(f"globals/{name}", value, self.step, dtype=dtype)

    def _dump_global_data(
        self, location: str, value: Any, step: int, dtype: str = None
    ) -> None:
        """Dump a global value / array to the file at location `location`.

        Args:
            location: Location in the file
            value: Data to dump
            dtype: Optional. Numpy style datatype, see h5py documentation,
                defaults to the dtype of the vector.
        """
        # split location into group and dataset
        group_name, dataset_name = location.rstrip("/").rsplit("/", 1)

        if isinstance(value, np.ndarray):
            shape = (step + 1, *value.shape)
        else:
            shape = (step + 1,)

        if self._prank == 0:
            with self._file("a") as f:
                grp = f.require_group(group_name)
                if dataset_name not in grp.keys():
                    vec = grp.create_dataset(
                        dataset_name,
                        shape=shape,
                        dtype=dtype if dtype else np.array(value).dtype,
                        chunks=True,
                        maxshape=(None, *shape[1:]) if len(shape) > 1 else (None,),
                        fillvalue=np.nan,
                    )
                    vec[-1] = value
                else:
                    vec = grp[dataset_name]
                    vec.resize(shape)
                    vec[-1] = value

    def add_global_fields(self, fields: Dict[str, Any]) -> None:
        """Add multiple global fields at once.

        Args:
            fields: Dictionary with fields
        """
        for name, value in fields.items():
            self.add_global_field(name, value)

    def finish_step(self) -> None:
        """Finish step. Adds 1 to the step counter."""
        self.step += 1

    def finish_sim(self, status: str = "Finished") -> None:
        if self._prank == 0:
            self.change_status(status)

    def register_git_attributes(self, repo_path: str = "./") -> None:
        """Register git information for given repo.

        Args:
            repo_path (`str`): path to git repository
        """
        if self._prank == 0:
            repo_path = os.path.abspath(repo_path)
            # store current working directory
            cwd = os.getcwd()

            # switch directory to git repo
            os.chdir(repo_path)
            git_string = GitStateGetter().create_git_string()

            # switch working directory back
            os.chdir(cwd)

            with self._file("a") as f:
                grp = f.require_group("git")
                repo_name = os.path.split(repo_path)[1]
                log.info(f"Adding repo {repo_name}")
                if repo_name in grp.keys():
                    del grp[repo_name]
                grp.create_dataset(repo_name, data=git_string)

    def copy_executable(self, script_path: str) -> None:
        """WILL BE REMOVED. USE COPY_FILE.
        Copy an executable to directory for reproducability.

        Args:
            script_path: path to script
        """
        shutil.copy(script_path, self.path)
        self.executable = os.path.split(script_path)[1]

    def copy_file(self, source: Union[str, list], destination: str = "") -> None:
        """Copy a file to the datafolder.

        Args:
            source: path to file, or list of files
            destination: destination (will create intermediatory directories)
        """
        if self._prank != 0:
            return

        if isinstance(source, list):
            for item in source:
                self.copy_file(item, destination)
            return

        destination = os.path.join(self.path, destination)

        if os.path.isdir(source):
            shutil.copytree(
                source,
                os.path.join(destination, os.path.basename(source)),
                dirs_exist_ok=True,
            )
        elif os.path.isfile(source):
            os.makedirs(destination, exist_ok=True)
            shutil.copy(source, destination)
        else:
            raise FileNotFoundError



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2024 Flavio Lorez and contributors
#
# There is no warranty for this code


class MockMPI:
    """
    Mock class for `mpi4py.MPI` to be used when MPI is not available or usage
    not desired. Not importing MPI increases launch speed significantly, which
    is important for CLI applications.

    Attributes:
        Comm: Mock class for `mpi4py.MPI.Comm`
        COMM_WORLD: Mock object for `mpi4py.MPI.COMM_WORLD`
        COMM_SELF: Mock object for `mpi4py.MPI.COMM_SELF`
        COMM_NULL: Mock object for `mpi4py.MPI.COMM_NULL`
    """

    class Comm:
        def __init__(self):
            self.size = 1
            self.rank = 0
            self.comm = None
            self.is_mpi = False
            self.is_master = True

        def barrier(self):
            pass

        def finalize(self):
            pass

        def bcast(self, data, root=0):
            return data

        def scatter(self, data, root=0):
            return data

        def gather(self, data, root=0):
            return data

        def allgather(self, data):
            return [data]

        def allreduce(self, data, op):
            return data

        def reduce(self, data, op, root=0):
            return data

        def send(self, data, dest, tag=0):
            pass

        def recv(self, source, tag=0):
            return None

        def recv_any_source(self, tag=0):
            return None

        def sendrecv(self, send_data, dest, tag=0):
            return None

        def sendrecv_replace(self, data, dest, tag=0):
            return data

        def get_processor_name(self):
            return "localhost"

        def get_version(self):
            return "0.0.0"

        def get_library_version(self):
            return "0.0.0"

        def get_error_string(self, errorcode):
            return "No error"

        def get_exception_class(self):
            return Exception

        def get_exception_string(self, errorcode):
            return "No error"

        def get_count(self, status, datatype):
            return 0

        def get_status(self, request):
            return None

        def get_source(self, status):
            return 0

        def get_tag(self, status):
            return 0

        def get_elements(self, status):
            return 0

        def get_bytes(self, status):
            return 0

        def get_cancelled(self, status):
            return False

        def get_topo(self):
            return None

        def get_cart(self):
            return None

        def get_dims(self):
            return None

        def get_coords(self):
            return None

        def get_rank(self):
            return 0

        def get_size(self):
            return 1

        def Get_size(self):
            return 1

    COMM_WORLD = Comm()
    COMM_SELF = Comm()
    COMM_NULL = Comm()



import sys

from bamboost import Manager


def create_test_run(
    db: Manager, nb_processes: int = 1, array_size: int = 10000, nb_steps: int = 100
) -> None:
    """Create a test run for given number of processes and array sizes."""
    params = {
        "nb_processes": nb_processes,
        "array_size": (array_size, array_size),
        "nb_steps": nb_steps,
    }
    script_file = "script.py"
    sim = db.create_simulation(f"{array_size}_{nb_processes:02d}", parameters=params, skip_duplicate_check=True)
    mpicommand = "" if nb_processes == 1 else f"mpirun -n {nb_processes}"

    commands = [
        f"{mpicommand} python3 $SIMULATION_DIR/{script_file} --path $SIMULATION_DIR/.. --uid {sim.uid}"
    ]
    sim.create_run_script(commands, euler=False, sbatch_kwargs={"--ntasks": nb_processes})
    sim.copy_file(script_file)


def main(test_manager_name: str = "out"):
    manager = Manager(test_manager_name)

    for nb_processes in [1, 2, 4, 8]:
        array_size = 20000
        nb_steps = 1
        create_test_run(manager, nb_processes, array_size, nb_steps)


if __name__ == "__main__":
    test_manager_name = sys.argv[1]
    main()



class ExtensionsLazyLoader:
    # This pattern is used to provide a single point/namespace to access all
    # the extensions without importing them directly.
    # 
    # TODO: this is the best way I could come up with, please suggest a better
    # approach if you have one.

    @property
    def FenicsWriter(self):
        from .fenics import FenicsWriter
        return FenicsWriter

    @property
    def Remote(self):
        from .remote_manager import Remote
        return Remote

    @property
    def RemoteManager(self):
        from .remote_manager import RemoteManager
        return RemoteManager

    @property
    def install_slurm(self):
        from .slurm import install
        return install


extensions = ExtensionsLazyLoader()



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

from __future__ import annotations

import os
import pkgutil
import sqlite3
import subprocess
from functools import cached_property, wraps
from typing import Callable

import pandas as pd

from bamboost import BAMBOOST_LOGGER
from bamboost._config import config
from bamboost._sqlite_database import SQLiteHandler, with_connection
from bamboost.common.mpi import MPI
from bamboost.common.utilities import unflatten_dict
from bamboost.index import DatabaseTable, IndexAPI
from bamboost.manager import Manager, ManagerFromUID
from bamboost.simulation import Simulation

__all__ = [
    "Remote",
    "RemoteManager",
    "RemoteSimulation",
]

log = BAMBOOST_LOGGER.getChild(__name__.split(".")[-1])

HOME = os.path.expanduser("~")
CACHE_DIR = os.path.join(HOME, ".cache", "bamboost")
# location off the index file on the remote server
REMOTE_INDEX = "~/.local/share/bamboost/bamboost.db"


# MonkeyPatch manager.ManagerFromUID
def _extend_manager_from_uid_getitem(original_getitem: Callable):
    """Extend the __getitem__ method of ManagerFromUID to handle remote keys.
    In the following format: ssh://<remote_name>/<id>.
    """

    @wraps(original_getitem)
    def modified_getitem(self: ManagerFromUID, key: str):
        # If key starts with ssh://, it is a remote key
        # Format: ssh://<remote_name>/<id>
        if key.startswith("ssh://"):
            remote_name = key.split("/")[2]
            key = key.split("/")[3]
            try:
                remote = Remote(remote_name, skip_update=True)
                return remote[key]
            except KeyError:
                remote = Remote(remote_name, skip_update=False)
                return remote[key]

        return original_getitem(self, key)

    return modified_getitem


# MonkeyPatch ManagerFromUID if not already patched
if not hasattr(ManagerFromUID.__getitem__, "__wrapped__"):
    ManagerFromUID.__getitem__ = _extend_manager_from_uid_getitem(
        ManagerFromUID.__getitem__
    )
    Manager.fromUID = ManagerFromUID()


# MonkeyPatch Simulation.fromUID
def _extend_simulation_from_uid(original_from_uid: Callable):
    """Extend the fromUID method of Simulation to handle remote keys.
    In the following format: ssh://<remote_name>/<id>:<uid>.
    """

    @wraps(original_from_uid)
    def modified_from_uid(cls: Simulation, key: str):
        # If key starts with ssh://, it is a remote key
        # Format: ssh://<remote_name>/<db_id>:<sim_id>
        if key.startswith("ssh://"):
            remote_name = key.split("/")[2]
            full_id = key.split("/")[3]
            db_id, sim_id = full_id.split(":")
            remote = Remote(remote_name, skip_update=True)
            return remote[db_id].sim(sim_id)

        return original_from_uid(key)

    return modified_from_uid


# MonkeyPatch Simulation if not already patched
if not hasattr(Simulation.fromUID, "__wrapped__"):
    Simulation.fromUID = classmethod(_extend_simulation_from_uid(Simulation.fromUID))


class Remote(IndexAPI, SQLiteHandler):
    """Access bamboost database of a remote server. The index is fetched using
    rsync over ssh. The `remote_name` can be a hostname or an IP address. Make
    sure that ssh keys are set and working, as there is no user authentication.
    The `skip_update` flag can be set to avoid fetching the index from the
    remote server.

    Args:
        - remote_name (str): The hostname or IP address of the remote server.
        - skip_update (bool): Flag to avoid fetching the index from the remote
          server. Default is False.

    Example:
        >>> remote = Remote("euler")
        >>> remote.read_table()
        returns a pandas DataFrame of the remote index.
        >>> remote["<id>"]
        returns a RemoteManager object for the given id.
    """

    def __new__(cls, *args, **kwargs):
        """Override the __new__ method to avoid the singleton pattern of IndexAPI."""
        return object.__new__(cls)

    def __init__(
        self, remote_name: str, skip_update: bool = False, *, home_path: str = None
    ) -> None:
        self.remote_name = remote_name
        self.local_path = os.path.join(CACHE_DIR, self.remote_name)
        os.makedirs(self.local_path, exist_ok=True)
        self.file = f"{self.local_path}/bamboost.db"

        if not skip_update:
            process = self.fetch_index()
            process.wait()

        # Initialize the SQLiteHandler
        SQLiteHandler.__init__(self, self.file)

    @classmethod
    def list(cls) -> list:
        """List all remote servers."""
        return [
            name
            for name in os.listdir(CACHE_DIR)
            if os.path.isdir(os.path.join(CACHE_DIR, name))
        ]

    def _ipython_key_completions_(self) -> list[str]:
        ids = self.read_table()[["id", "path"]].values
        completion_keys = [
            f'{key} - {"..."+val[-25:] if len(val)>=25 else val}' for key, val in ids
        ]
        return completion_keys

    def __getitem__(self, id: str) -> RemoteManager:
        """Return a `RemoteManager` for the given id."""
        id = id.split(" - ")[0]
        return RemoteManager(id, remote=self)

    def fetch_index(self) -> subprocess.Popen:
        """Fetch the index from the remote server."""
        return subprocess.Popen(
            ["rsync", "-av", f"{self.remote_name}:{REMOTE_INDEX}", self.file],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )

    def get_manager(self, id: str, skip_update: bool = False) -> RemoteManager:
        return RemoteManager(id, remote=self, skip_update=skip_update)

    @with_connection
    def get_path(self, id: str) -> str:
        self._cursor.execute("SELECT path FROM dbindex WHERE id=?", (id,))
        fetch = self._cursor.fetchone()
        if fetch is None:
            raise KeyError(f"No database found with id: {id}")
        else:
            return fetch[0]

    @with_connection
    def insert_local_path(self, id: str, path: str) -> None:
        try:
            self._cursor.execute(
                "UPDATE dbindex SET local_path=? WHERE id=?", (path, id)
            )
        except sqlite3.OperationalError:
            self._cursor.execute(
                "ALTER TABLE dbindex ADD COLUMN local_path TEXT DEFAULT NULL"
            )
            self._cursor.execute(
                "UPDATE dbindex SET local_path=? WHERE id=?", (path, id)
            )


class RemoteDatabaseTable(DatabaseTable):
    def sync(self) -> None:
        """Don't sync a remote database."""
        return None


class RemoteManager(Manager):
    """
    Manager class with remote functionality. Constructor takes an existing ID
    of a database on a remote server. The ssh connection must be set up to work
    without explicit user authentication. Data is lazily transferred using
    rsync.
    """

    # INFO: The initialization of this class takes around 8 seconds because the
    # home directory of the remote server needs to be fetched, read database
    # index, and get remote database csv file with db.df in it (see below).
    def __init__(
        self,
        id: str,
        remote: Remote,
        comm: MPI.Comm = MPI.COMM_WORLD,
    ) -> None:
        """
        params
        skip_update: if True, does not lookup the new database on the remote:
        """
        self.UID = id
        self.remote = remote
        self.comm = comm
        self.path = os.path.join(self.remote.local_path, self.UID)
        log.info(f"Creating cache directory at {self.path}")
        os.makedirs(self.path, exist_ok=True)

        self.remote_path_db = self._index.get_path(self.UID)

        # check if path exists
        if not os.path.isdir(self.path):
            os.makedirs(self.path)

        # Write the database ID file if it does not exist
        if not os.path.exists(f"{self.path}/.BAMBOOST-{self.UID}"):
            with open(f"{self.path}/.BAMBOOST-{self.UID}", "w") as f:
                f.write(self.UID)

        self.UID = id
        self._index.insert_local_path(self.UID, self.path)

        # Update the SQL table for the database
        with self._index.open():
            self._table.create_database_table()
            # self._table.sync()

    def _repr_html_(self) -> str:
        html_string = pkgutil.get_data(__name__, "../html/manager.html").decode()
        icon = pkgutil.get_data(__name__, "../html/icon.txt").decode()
        return (
            html_string.replace("$ICON", icon)
            .replace("$db_path", self.path)
            .replace("$db_uid", "euler-" + self.UID)
            .replace("$db_size", str(len(self)))
        )

    def _get_uids(self) -> list[str]:
        """Override the simulation list to fetch from databasetable."""
        uids_in_tuple = self._index.fetch(f"SELECT id FROM db_{self.UID}")
        return [uid for (uid,) in uids_in_tuple]

    @property
    def _index(self) -> IndexAPI:
        return self.remote

    @property
    def _table(self) -> RemoteDatabaseTable:
        return RemoteDatabaseTable(self.UID, _index=self._index)

    def get_view(self, include_linked_sims: bool = False) -> pd.DataFrame:
        df = super().get_view(include_linked_sims)
        df.insert(1, "cached", False)
        for id in os.listdir(self.path):
            if id in df["id"].values:
                df.loc[df["id"] == id, "cached"] = True

        opts = config.get("options", {})
        if "sort_table_key" in opts:
            df.sort_values(
                opts.get("sort_table_key", "id"),
                ascending=opts.get("sort_table_order", "asc") == "asc",
                inplace=True,
            )
        return df

    def sim(self, uid, return_writer: bool = False):
        """Return simulation object.

        Args:
            - uid (str): the unique id of the sim to be transferred.
            - return_writer (bool): Flag to indicate whether to return a writer
              object. Default is False.

        This method checks if the data for the given uid is already in the
        local cache. If not, it transfers the data from a remote location using
        rsync. The method then calls the superclass method to perform further
        operations on the transferred data.
        """
        # Check if data is already in cache
        if os.path.exists(f"{self.path}/{uid}"):
            log.info(f"Data for {uid} already in cache")
            return RemoteSimulation(uid, self)

        # Transfer data using rsync
        log.info(f"Data not in cache. Transferring data for {uid} from {self.remote}")
        self.rsync(uid)

        return RemoteSimulation(uid, self)

    def shallow_sim(self, uid: str) -> RemoteSimulation:
        """Return a RemoteSimulation object without transferring the data.

        Args:
            uid: The unique id of the simulation.
        """
        return RemoteSimulation(uid, self, _shallow=True)

    def _rsync(self, uid: str | None = None) -> subprocess.Popen:
        """Transfer data using rsync. This method is called by the `rsync`.
        It returns the subprocess.Popen object.

        Args:
            uid: The unique id of the simulation to be transferred. If None,
                all simulations are synced.
        """
        if uid is not None:
            log.info(f"Start syncing data for {uid} with {self.path}")
            return subprocess.Popen(
                [
                    "rsync",
                    "-ravh",
                    f"{self.remote.remote_name}:{self.remote_path_db}/{uid}",
                    f"{self.path}",
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
            )

        # else, sync entire database directory
        log.info(f"Start sync database with {self.path}")
        return subprocess.Popen(
            [
                "rsync",
                "-ravh",
                f"{self.remote.remote_name}:{self.remote_path_db}/*",
                f"{self.path}",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )

    def rsync(
        self, uid: str | None = None
    ) -> RemoteManager:
        """Transfer data using rsync. Wait for the process to finish and return
        self.

        Args:
            uid: The unique id of the simulation to be transferred. If None,
                all simulations are synced.
        """
        process = self._rsync(uid)
        for line in iter(process.stdout.readline, ""):
            print(line, end="")
        process.wait()
        return self


class RemoteSimulation(Simulation):
    cached: bool = False
    shallow: bool = False

    def __init__(
        self, uid: str, manager: RemoteManager, *, _shallow: bool = False
    ) -> None:
        self.manager = manager
        self.uid = uid

        if not _shallow:
            self._init_base_class()
        else:
            # A subset of the Simulation __init__ method
            self.uid: str = uid
            self.path_database: str = os.path.abspath(manager.path)
            self.database_id = manager.UID
            self.path: str = os.path.abspath(os.path.join(manager.path, uid))
            self.h5file: str = os.path.join(self.path, f"{self.uid}.h5")
            self.xdmffile: str = os.path.join(self.path, f"{self.uid}.xdmf")

            # set flag to indicate that the data is not transferred
            self.shallow = True

    @cached_property
    def parameters(self) -> dict:
        """Returns the parameters of the simulation.

        - If the simulation is shallow, the parameters are fetched from the
          database. If so, the metadata is included in the parameters.
        - If the simulation is not shallow, the parameters are fetched from the
          HDF5 file.
        """
        if self.shallow:
            return unflatten_dict(self.manager._table.read_entry(self.uid).to_dict())
        return super().parameters

    def _init_base_class(self) -> None:
        super().__init__(self.uid, self.manager.path, _db_id=self.manager.UID)
        self.cached = True

    def sync(self) -> RemoteSimulation:
        """Sync the simulation data with the remote server.

        Args:
            stdout: The file object to write the output of the rsync process.
        """
        self.manager.rsync(self.uid)
        self._init_base_class()
        return self

    def get_full_uid(self) -> str:
        return f"ssh://{self.manager.remote.remote_name}/{self.manager.UID}:{self.uid}"


if __name__ == "__main__":
    pass



# test_manager.py

import os
import shutil
import tempfile

import numpy as np
import pandas as pd
import pytest

from bamboost import Manager, Simulation, SimulationWriter, index


@pytest.fixture()
def temp_manager():
    temp_dir = tempfile.mkdtemp()
    db = Manager(path=temp_dir)
    yield db
    try:
        shutil.rmtree(temp_dir)
    except FileNotFoundError:
        pass


def test_if_new_manager_created():
    temp_dir = tempfile.mkdtemp()
    db = Manager(temp_dir)
    assert os.path.isdir(temp_dir)


class TestManager:
    def test_create_simulation_creates_new_file(self, temp_manager: Manager):
        sim_uid = "this"
        writer = temp_manager.create_simulation(uid=sim_uid)
        h5_file = os.path.join(temp_manager.path, sim_uid, f"{sim_uid}.h5")
        assert os.path.isfile(h5_file)

    @pytest.mark.parametrize("prefix", ["dummy"])
    def test_create_simulation_argument_prefix(self, temp_manager: Manager, prefix):
        sim_uid = "this"
        writer = temp_manager.create_simulation(uid=sim_uid, prefix=prefix)

        assert writer.uid == f"{prefix}_this"

    def test_create_simulation_argument_note(self, temp_manager: Manager):
        sim_uid = "this"
        writer = temp_manager.create_simulation(uid=sim_uid, note="dummy note")

        assert writer.metadata["notes"] == "dummy note"

    def test_create_simulation_argument_files(self, temp_manager: Manager):
        sim_uid = "this"
        # Create Files
        with open(os.path.join(temp_manager.path, "a.txt"), "w") as f:
            f.write("a")
        with open(os.path.join(temp_manager.path, "b.txt"), "w") as f:
            f.write("b")

        files = [f"{temp_manager.path}/a.txt", f"{temp_manager.path}/b.txt"]
        writer = temp_manager.create_simulation(uid=sim_uid, files=files)

        assert all([os.path.isfile(os.path.join(writer.path, f)) for f in files])

    def test_create_simulation_argument_links(self, temp_manager: Manager):
        sim_uid = "this"
        linked_sim = temp_manager.create_simulation(uid="link_to_this")
        writer = temp_manager.create_simulation(
            uid=sim_uid, links={"mesh": linked_sim.get_full_uid()}
        )

        assert writer.links["mesh"].metadata == linked_sim.metadata

    @pytest.mark.parametrize("fix_df", [True, False])
    def test_manager_length(self, temp_manager: Manager, fix_df: bool):
        temp_manager.FIX_DF = fix_df
        temp_manager.create_simulation()
        temp_manager.create_simulation()
        temp_manager.create_simulation()
        assert len(temp_manager) == 3

    def test_get_single_simulation_base_case(self, temp_manager: Manager):
        sim_uid = "this"
        temp_manager.create_simulation()
        temp_manager.create_simulation(sim_uid)
        db = Manager(temp_manager.path)
        sim = db.sim(sim_uid)
        assert isinstance(sim, Simulation)
        assert sim.path == os.path.join(temp_manager.path, sim_uid)


class TestManagerDuplicate:
    def test_prompt(self, monkeypatch, temp_manager: Manager):
        def patchinput(*args, **kwargs):
            raise RuntimeError("input called")

        monkeypatch.setattr("builtins.input", patchinput)
        with pytest.raises(RuntimeError, match="input called"):
            params = dict(a=1, b=2)
            temp_manager.create_simulation(parameters=params)
            sim = temp_manager.create_simulation(
                parameters=params, duplicate_action="prompt"
            )

        # Check that there is no prompt here
        params = dict(a=1, b=2)
        sim = temp_manager.create_simulation(parameters=params, duplicate_action="a")

    def test_abort(self, temp_manager: Manager):
        params = dict(a=1, b=2)
        temp_manager.create_simulation(parameters=params)
        sim = temp_manager.create_simulation(parameters=params, duplicate_action="a")
        assert sim is None
        assert len(temp_manager) == 1

    def test_create_new(self, temp_manager: Manager):
        params = dict(a=1, b=2)
        temp_manager.create_simulation(parameters=params)
        sim = temp_manager.create_simulation(parameters=params, duplicate_action="c")
        assert isinstance(sim, SimulationWriter)
        assert len(temp_manager) == 2


# --------------------------------------------------------------------------
# Database syncing between SQLite and Filesystem
# --------------------------------------------------------------------------
def test_if_table_created(temp_manager: Manager):
    uid = temp_manager.UID
    table_name = f"db_{uid}"
    res = index.IndexAPI().fetch("SELECT name FROM sqlite_master WHERE type='table';")
    # assert that the database table exists
    assert (table_name,) in res
    # assert that the table with update times exists
    assert (f"db_{uid}_t",) in res


def test_dataframe_integrity(temp_manager: Manager):
    _index = index.IndexAPI()  # shortcutting calling IndexAPI() every time
    booleans = {
        "1": True,
        "2": False,
        "3": True,
    }

    @_index.commit_once
    def create_sims(booleans: list):
        for args in zip([1, 2, 3], booleans, ["a", "b", "c"]):
            sim = temp_manager.create_simulation(
                uid=args[1],
                parameters={
                    "int": args[0],
                    "float": 1.0,
                    "str": args[2],
                    "boolean": booleans[args[1]],
                    "boolean2": False,
                    "array": np.array([1, 2, 3]),
                },
                skip_duplicate_check=True,
            )

    create_sims(booleans)

    # delete the table to force re-creation -> booleans are wrong!!!
    with _index.open():
        _index._cursor.execute(f"DROP TABLE db_{temp_manager.UID};")
        _index._cursor.execute(f"DROP TABLE db_{temp_manager.UID}_t;")
        _index.commit()

    db = Manager(uid=temp_manager.UID)
    df = db.df.set_index("id")

    for uid, val in booleans.items():
        assert df.loc[uid, "boolean"] == val


def test_duplicates_detect(temp_manager: Manager):
    db = temp_manager
    params1 = dict(a=1, b=2)
    db.create_simulation("1", params1)
    db.create_simulation("2", dict(a=4, b=5))

    assert db._list_duplicates(params1)[0] == "1"

    sim = db.create_simulation(parameters=params1, duplicate_action="c")

    assert {"1", sim.uid} == set(db._list_duplicates(params1))


def test_duplicates_new_key(temp_manager: Manager):
    db = temp_manager
    params1 = dict(a=1, b=2)
    db.create_simulation("1", params1)

    params2 = dict(a=1, b=2, c=3)

    assert len(db._list_duplicates(params2)) == 0


def test_duplicates_handles_nan(temp_manager: Manager):
    db = temp_manager
    params1 = dict(a=1, b=2)
    db.create_simulation("1", params1)

    params2 = dict(a=1, b=2, c=3)
    db.create_simulation("2", params2)
    # This introduces a nan entry in simulation "1"
    assert np.isnan(db.df[db.df.id == "1"].c.iloc[0])

    params3 = dict(a=1, b=2, c=4)
    # Simulation 2 shall not be regarded as duplicate of simulation 1, because of the NaN
    assert len(db._list_duplicates(params3)) == 0


def test_duplicates_lists(temp_manager: Manager):
    db = temp_manager
    params1 = dict(a=[1, 2, 3])
    db.create_simulation("1", params1)

    params2 = dict(a=[2, 3, 4])
    assert len(db._list_duplicates(params2)) == 0

    db.create_simulation("2", params2)

    assert len(db._list_duplicates(params2)) == 1
    assert len(db._list_duplicates(params1)) == 1

    # Works with arrays
    params2 = dict(a=np.array([2, 3, 4]))
    assert len(db._list_duplicates(params2)) == 1
    db.create_simulation("3", params2, duplicate_action="c")
    assert len(db._list_duplicates(params2)) == 2

    # different length -> dnot duplicates
    # Works with arrays
    params2 = dict(a=np.array([2, 3, 4, 5]))
    assert len(db._list_duplicates(params2)) == 0

    # 2D arrays
    params2 = dict(a=np.array([[2, 3], [4, 5]]))
    assert len(db._list_duplicates(params2)) == 0
    db.create_simulation(parameters=params2, duplicate_action="c")
    assert len(db._list_duplicates(params2)) == 1

    # We do not plan to support sets


def test_replace_duplicates(temp_manager: Manager):
    db = temp_manager
    params1 = dict(a=[1, 2, 3])
    db.create_simulation("1", parameters=params1)
    assert len(db.df) == 1

    db.create_simulation("2", parameters=params1, duplicate_action="r")
    assert len(db.df) == 1
    assert db._get_uids()[0] == "2"


def test_altered_uid(temp_manager: Manager):
    db = temp_manager
    params1 = dict(a=[1, 2, 3])
    db.create_simulation("1", parameters=params1)
    assert len(db.df) == 1

    db.create_simulation(parameters=params1, duplicate_action="c")
    db.get_view()

    assert len(db.df) == 2

    assert set(db.df["id"]) == {"1", "1.1"}


@pytest.mark.parametrize(
    "params,expected",
    [
        (dict(a=1, b=2), ("1", "2", "3")),
        (dict(a=1, b=2, c=3), ("2",)),
        (dict(c=lambda c: c < 4), ("2",)),
    ],
)
def test_find(temp_manager: Manager, params: dict, expected: tuple):
    db = temp_manager
    # mock the database dataframe from sql
    db._dataframe = pd.DataFrame.from_records(
        [
            dict(id="1", a=1, b=2),
            dict(id="2", a=1, b=2, c=3),
            dict(id="3", a=1, b=2, c=4),
            dict(id="4", a=1, b=10),
        ]
    )

    assert isinstance(db.find(params), pd.DataFrame)
    assert set(db.find(params).id) == set(expected)



import numpy as np
from scipy import spatial
from itertools import combinations
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection


def plot_lattice(coords, connectivity, displacements, ax=None):
    if ax is None:
        ax = plt.gca()
    vec = coords + displacements
    lines = LineCollection(vec[connectivity], color='.7', linewidth=.7)
    ax.add_collection(lines)
    ax.set_aspect('equal')
    ax.set_ylim(-.1, 1.1)
    ax.set_xlim(-.1, np.max(vec[:, 0])+.1)
    return ax


def get_rot_tensor(cos, sin):
    R = np.array([
        [cos, sin, 0, 0],
        [-sin, cos, 0, 0],
        [0, 0, cos, sin],
        [0, 0, -sin, cos]
        ])
    return R


class TrussElement:

    def __init__(self, nodes, coords, E, A) -> None:
        self.i, self.j = nodes
        self.coords = coords  # current position vector
        self.E = E
        self.A = A

    def get_truss_matrix_local(self, L) -> np.ndarray:
        self.k = self.E*self.A/L * np.array([
            [1, 0, -1, 0],
            [0, 0, 0, 0],
            [-1, 0, 1, 0],
            [0, 0, 0, 0]
            ])
        return self.k

    def get_matrix_global(self) -> np.ndarray:
        xi = self.coords[self.i, :]
        xj = self.coords[self.j, :]
        L = np.linalg.norm(xi-xj)
        cos, sin = (xi[0]-xj[0])/L, (xi[1]-xj[1])/L

        k_local = self.get_truss_matrix_local(L)
        R = get_rot_tensor(cos, sin)
        k_global = R.T.dot(k_local).dot(R)
        return k_global


class LatticeModel:

    def __init__(self, E, N: int = None, randomness: float = None,
                 step: float = None) -> None:
        self.E = E
        self.bcs = None
        if N is not None:
            self.create_structure(N)
        if randomness is not None:
            self.add_randomness(randomness)
        if step is not None:
            self.move_top(step)
        
    def create_structure(self, N):
        xi = np.linspace(0, 1, N)
        yi = np.linspace(0, 1, N)
        xx, yy = np.meshgrid(xi, yi)
        self.N = N
        self.coordinates = np.vstack((xx.flatten(), yy.flatten())).T
        self.nb_dofs = self.coordinates.size
        self.top_indices = np.where(self.coordinates[:, 1]==1)[0]
        self.bottom_indices = np.where(self.coordinates[:, 1]==0)[0]
        triangulation = spatial.Delaunay(self.coordinates)
        connectivity = list()
        for el in triangulation.simplices:
            connectivity.extend((edge for edge in combinations(el, 2)))
        self.connectivity = np.unique(np.array(connectivity), axis=0)

        # assign area such that total L*A = 1
        self.lengths = np.linalg.norm(self.coordinates[self.connectivity, 0]-self.coordinates[self.connectivity, 1], axis=1)
        # self.areas = np.full(self.connectivity.shape[0], 1) / np.sum(self.lengths)
        self.areas = np.full(self.connectivity.shape[0], 1/self.connectivity.shape[0])
        # create matrices
        self.U = np.zeros((self.nb_dofs, 1))
        self.U_history = np.zeros((self.nb_dofs, 1))
        self.F = np.zeros((self.nb_dofs, 1))
        self.F_history = np.zeros((self.nb_dofs, 1))
        self.K = np.zeros((self.nb_dofs, self.nb_dofs))

        return self

    def add_randomness(self, scale: float) -> None:
        self.coordinates += np.random.normal(0, scale, self.coordinates.shape) * 1/self.N
        return self

    def add_bcs(self, bcs: list) -> None:
        self.bcs = bcs
        return self

    def move_top(self, value: float) -> None:
        u_top = np.vstack((
            self.top_indices,
            np.full(self.top_indices.shape, value),
            np.full(self.top_indices.shape, 0))
                          )
        u_bottom = np.vstack((
            self.bottom_indices,
            np.full(self.bottom_indices.shape, 0),
            np.full(self.bottom_indices.shape, 0))
                             )
        self.add_bcs(np.vstack((u_top.T, u_bottom.T)))
        return self

    def end_step(self) -> None:
        self.coordinates += self.U.reshape(-1, 2)
        self.U_history += self.U
        self.U = np.zeros((self.nb_dofs, 1))
        self.F_history += self.F
        self.F = np.zeros((self.nb_dofs, 1))
        return self

    @property
    def displacements(self):
        return (self.U_history + self.U).reshape(-1, 2)

    @property
    def forces(self):
        return (self.F_history + self.F).reshape(-1, 2)
    
    def get_force_bottom(self):
        return self.forces[self.bottom_indices].sum(axis=0)

    def assemble(self) -> None:

        for el_nb, (i, j) in enumerate(self.connectivity):
            element = TrussElement((i, j), self.coordinates, self.E, self.areas[el_nb])
            k_local = element.get_matrix_global()

            self.K[2*i:2*i+2, 2*i:2*i+2] += k_local[0:2, 0:2]
            self.K[2*i:2*i+2, 2*j:2*j+2] += k_local[0:2, 2:4]
            self.K[2*j:2*j+2, 2*i:2*i+2] += k_local[2:4, 0:2]
            self.K[2*j:2*j+2, 2*j:2*j+2] += k_local[2:4, 2:4]

        self.zerorows = np.where(np.all(self.K==0, axis=1))[0]
        return self

    def apply_bcs(self) -> None:
        all_dofs = np.arange(0, self.nb_dofs)
        restrained_dofs = list()
        for (node, ux, uy) in self.bcs:
            node = int(node)
            restrained_dofs.extend([2*node, 2*node+1])
            self.U[2*node] = ux
            self.U[2*node+1] = uy

        free_dofs = np.setdiff1d(all_dofs, restrained_dofs)

        self.Kff = self.K[free_dofs, :][:, free_dofs]
        self.Kfr = self.K[free_dofs, :][:, restrained_dofs]
        self.Krf = self.K[restrained_dofs, :][:, free_dofs]
        self.Krr = self.K[restrained_dofs, :][:, restrained_dofs]

        self.Ff = self.F[free_dofs, :]
        self.Fr = self.F[restrained_dofs, :]
        self.Uf = self.U[free_dofs, :]
        self.Ur = self.U[restrained_dofs, :]

        zerorows_free = np.where(np.all(self.Kff==0, axis=1))[0]
        self.zerorows_free = zerorows_free
        self.Kff = np.delete(self.Kff, zerorows_free, axis=0)
        self.Kff = np.delete(self.Kff, zerorows_free, axis=1)

        self.free_dofs_notzero = np.delete(free_dofs, zerorows_free, axis=0)
        self.Ff = np.delete(self.Ff, zerorows_free, axis=0)

    def solve(self) -> None:
        self.K = np.zeros((self.nb_dofs, self.nb_dofs))
        self.U = np.zeros((self.nb_dofs, 1))
        self.F = np.zeros((self.nb_dofs, 1))

        self.assemble()
        self.apply_bcs()

        self.Uf = np.linalg.solve(self.Kff, self.Ff - self.Kfr.dot(self.Ur))
        self.U[self.free_dofs_notzero] = self.Uf
        self.F = self.K.dot(self.U)
        
        self.end_step()  # updates the coordinates, adds disp to history

        return self

    def plot(self, vector=None, ax: plt.Axes = None):
        if ax is None:
            ax = plt.gca()
        if vector is None:
            vector = self.coordinates

        lines = LineCollection(vector[self.connectivity], color='.7', linewidth=.7)
        ax.add_collection(lines)
        ax.set_aspect('equal')
        ax.set_ylim(-.1, 1.1)
        ax.set_xlim(-.1, np.max(vector[:, 0])+.1)
            
    def plot_deformed(self, ax=None):
        self.plot(self.coordinates + self.U.reshape(-1, 2), ax)




# test index.py

import os
import shutil
import sqlite3

import numpy as np
import pytest

from bamboost.index import DatabaseTable, IndexAPI
from bamboost.manager import Manager

from test_manager import temp_manager

# --------------------------------------------------------------------------
# Index creation and database addition (SQLite)
# The index.Index is rerouted to a temporary file -> empty at the beginning
# --------------------------------------------------------------------------
def test_if_index_created():
    assert os.path.isfile(IndexAPI().file)


def test_if_table_created():
    conn = sqlite3.connect(IndexAPI().file, detect_types=sqlite3.PARSE_DECLTYPES)
    cursor = conn.cursor()
    # check if table dbindex exists
    cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='dbindex';"
    )
    assert cursor.fetchone() is not None


def test_if_database_added(temp_manager: Manager):
    # check if database is in index
    index = IndexAPI().read_table()  # returns pd.DataFrame
    assert temp_manager.UID in index.id.values
    assert temp_manager.path == index.loc[index.id == temp_manager.UID].path.values[0]


def test_if_database_removed(temp_manager: Manager):
    # check if database is removed from index if deleted
    shutil.rmtree(temp_manager.path)
    index = IndexAPI().clean().read_table()
    assert temp_manager.UID not in index.id.values


def test_index_clean(temp_manager: Manager):
    # check if index is cleaned
    shutil.rmtree(temp_manager.path)
    IndexAPI().clean()
    index = IndexAPI().read_table()
    assert index.empty


def test_index_clean_purge(temp_manager: Manager):
    # check if index is cleaned and purged (=table & entry in index is deleted)
    # shutil.rmtree(temp_manager.path)
    uid = temp_manager.UID
    table_name = f"db_{uid}"
    res = IndexAPI().fetch("SELECT name FROM sqlite_master WHERE type='table';")
    # assert that the database table exists
    assert (table_name,) in res

    shutil.rmtree(temp_manager.path)
    IndexAPI().clean(purge=True)
    res = IndexAPI().fetch("SELECT name FROM sqlite_master WHERE type='table';")
    assert f"db_{temp_manager.UID}" not in res


def test_getitem(temp_manager: Manager):
    # check if __getitem__ works -> returns index.DatabaseTable
    dbt = IndexAPI()[temp_manager.UID]
    assert isinstance(dbt, DatabaseTable)


def test_get_id_from_path(temp_manager: Manager):
    # check if get_id_from_path works
    assert IndexAPI().get_id(temp_manager.path) == temp_manager.UID


# --------------------------------------------------------------------------
# Correct transformation and retrieval of data types when writing to SQLite
# --------------------------------------------------------------------------
@pytest.mark.parametrize("value", [True, False, np.True_, np.False_])
def test_datatype_parsing_bool(temp_manager: Manager, value: bool):
    # check if boolean is parsed correctly
    sim = temp_manager.create_simulation(parameters={"test": value})
    series = DatabaseTable(temp_manager.UID).read_entry(sim.uid)
    assert series.test == value


@pytest.mark.parametrize("value", [1, 1.0])
def test_datatype_parsing_int_float(temp_manager: Manager, value):
    # check if datatype is parsed correctly
    sim = temp_manager.create_simulation(parameters={"test": value})
    series = DatabaseTable(temp_manager.UID).read_entry(sim.uid)
    assert type(series.test) == type(value)


@pytest.mark.parametrize("array", [np.array([[1, 2, 3], [4.0, 5.0, 6.0]])])
def test_datatype_parsing_array(temp_manager: Manager, array):
    # check if array is parsed correctly
    sim = temp_manager.create_simulation(parameters={"test": array})
    series = DatabaseTable(temp_manager.UID).read_entry(sim.uid)
    assert isinstance(series.test, np.ndarray)


def test_nested_dict(temp_manager: Manager):
    # check if nested dict is parsed correctly (dots are replaced by literal 'DOT' and replaced back when reading)
    sim = temp_manager.create_simulation(
        parameters={"test": {"a": 1, "b": 2, "c": {"d": np.array([2, 3, 4])}}}
    )
    series = DatabaseTable(temp_manager.UID).read_entry(sim.uid)
    assert series["test.a"] == 1
    assert series["test.b"] == 2
    assert np.array_equal(series["test.c.d"], np.array([2, 3, 4]))


def test_read_database_table(temp_manager: Manager):
    # check if database table is read correctly
    temp_manager.create_simulation(
        parameters={"test": True, "false": False, "int": 1, "float": 1.0}
    )
    table = DatabaseTable(temp_manager.UID).read_table()
    assert table.shape[0] == 1



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code
from __future__ import annotations

import os
import time
from functools import wraps
from typing import Any, Literal, Type

import h5py

from bamboost.common import mpi
from bamboost import BAMBOOST_LOGGER

log = BAMBOOST_LOGGER.getChild(__name__.split(".")[-1])

__all__ = [
    "open_h5file",
    "FileHandler",
    "with_file_open",
    "capture_key_error",
    "HAS_MPIO",
    "MPI_ACTIVE",
    "FILE_MODE_HIRARCHY",
]

HAS_MPIO = "mpio" in h5py.registered_drivers()
if HAS_MPIO and mpi.MPI_ON:
    MPI_ACTIVE = h5py.h5.get_config().mpi
else:
    MPI_ACTIVE = False

FILE_MODE_HIRARCHY = {
    "r": 1,
    "r+": 2,
    "a": 2,
    "w": 3,
}


def open_h5file(
    file: str,
    mode: Literal["mpio"] | Type[None],
    driver: bool | Type[None] = None,
    comm=None,
):
    """Open h5 file. Waiting if file is not available.

    Args:
        file (str): File to open
        mode (str): 'r', 'a', 'w', ...
        driver (str): driver for h5.File
        comm: MPI communicator
    """
    while True:
        try:
            if driver == "mpio" and MPI_ACTIVE and mpi.MPI_ON:
                return h5py.File(file, mode, driver=driver, comm=comm)
            else:
                return h5py.File(file, mode)

        except BlockingIOError:
            log.warning(f"file locked --> {file}")
            time.sleep(0.2)


def with_file_open(mode: str = "r", driver=None, comm=None):
    """Open the file (`self._file`) before function
    Close the file after the function call

    Works on classes containing the member `_file` of type :class:`~bamboost.common.file_handler.FileHandler`
    """

    def decorator(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            with self._file(mode, driver, comm):
                return method(self, *args, **kwargs)

        return wrapper

    return decorator


def capture_key_error(method):
    @wraps(method)
    def inner(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except KeyError as e:
            e.add_note(f"[file: {self.file_name}]")
            raise e

    return inner


class FileHandler:
    """File handler for an hdf5 file with the purpose of handling opening and closing
    of the file. We use the concept of composition to include an object of this type
    in classes which need access to an hdf5 file (such as the hdf5pointer and Simulation.)

    Args:
        file_name: the path to the file

    Attributes:
        file_object: the h5py file object (accessible if open)
        _lock: lock is kind of a stack. `open` increases the stack. `close` decreases
            the stack. file_object is only closed if the stack is at 0. Ensures consecutive
            method calls works. Would be a problem if the file is closed after each
            sub-operation.
        _mode: file mode
        _driver: file driver
        _comm: MPI communicator
    """

    def __init__(
        self, file_name: str, _comm: mpi.MPI.Comm = mpi.MPI.COMM_WORLD
    ) -> None:
        self.file_object: h5py.File = None
        self.file_name = file_name
        self.simulation_uid = os.path.basename(file_name)
        self._lock = 0
        self._mode = "r"
        self._driver = None
        self._comm = _comm

    def __call__(self, mode: str = "r", driver=None, comm=None) -> FileHandler:
        """Used to set the options for file opening.
        Example: `with sim._file('a', driver='mpio') as file:`
        """
        self._mode = mode
        self._driver = driver
        self._comm = comm if comm is not None else self._comm
        return self

    @capture_key_error
    def __getitem__(self, key) -> Any:
        return self.file_object[key]

    @capture_key_error
    def __delitem__(self, key) -> None:
        del self.file_object[key]

    @capture_key_error
    def __getattr__(self, __name: str) -> Any:
        try:
            return self.file_object.__getattribute__(__name)
        except AttributeError:
            return self.__getattribute__(__name)

    def __enter__(self):
        self.open(self._mode, self._driver, self._comm)
        return self

    def __exit__(self, *args):
        self.close()

    def open(self, mode: str = "r", driver=None, comm=None):
        if self._lock <= 0:
            log.debug(f"[{id(self)}] Open {self.file_name}")
            self.file_object = open_h5file(self.file_name, mode, driver, comm)

        if FILE_MODE_HIRARCHY[self.file_object.mode] < FILE_MODE_HIRARCHY[mode]:
            self.change_file_mode(mode, driver, comm)

        log.debug(f"[{id(self)}] Lock stack {self._lock}")
        self._lock += 1
        return self.file_object

    def close(self):
        self._lock -= 1
        if self._lock == 0:
            log.debug(f"[{id(self)}] Close {self.file_name}")
            self.file_object.close()
        log.debug(f"[{id(self)}] Lock stack {self._lock}")

    def change_file_mode(self, mode: str, driver=None, comm=None):
        log.info(
            f"Forced closing and reopening to change file mode [{self.file_name}]."
        )
        self.file_object.close()
        self.file_object = open_h5file(self.file_name, mode, driver, comm)



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

import os
import xml.etree.ElementTree as ET

from bamboost.common.file_handler import FileHandler

__all__ = ["XDMFWriter"]

# `numpy_to_xdmf_dtype` from `meshio.xdmf.common`
numpy_to_xdmf_dtype = {
    "int8": ("Int", "1"),
    "int16": ("Int", "2"),
    "int32": ("Int", "4"),
    "int64": ("Int", "8"),
    "uint8": ("UInt", "1"),
    "uint16": ("UInt", "2"),
    "uint32": ("UInt", "4"),
    "uint64": ("UInt", "8"),
    "float32": ("Float", "4"),
    "float64": ("Float", "8"),
}


class XDMFWriter:
    """Write xdmf file for a subset of the stored data in the H5 file.

    Args:
        filename (str): xdmf file path
        h5file (str): h5 file path"""

    def __init__(self, filename: str, _file: FileHandler):
        self.filename = filename
        self._file = _file
        self.h5file = os.path.basename(_file.file_name)
        self.xdmf_file = ET.Element("Xdmf", Version="3.0")
        self.domain = ET.SubElement(self.xdmf_file, "Domain")
        ET.register_namespace("xi", "https://www.w3.org/2001/XInclude/")
        self.mesh_name = "mesh"

    def write_file(self):
        tree = ET.ElementTree(self.xdmf_file)
        self._pretty_print(tree.getroot())
        tree.write(self.filename)

    def _pretty_print(self, elem, level=0):
        indent = "  "  # 4 spaces
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = "\n" + indent * (level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = "\n" + indent * level
            for elem in elem:
                self._pretty_print(elem, level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = "\n" + indent * level
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = "\n" + indent * level

    def write_points_cells(self, points_location: str, cells_location: str):
        """Write the mesh to the xdmf file.

        Args:
            points (str): String to geometry/nodes in h5 file
            cells (str): String to topology/cells in h5 file
        """
        grid = ET.SubElement(
            self.domain, "Grid", Name=self.mesh_name, GridType="Uniform"
        )
        self._points(grid, points_location)
        self._cells(grid, cells_location)

    def _points(self, grid: ET.Element, points_location: str):
        geometry_type = "XY"

        with self._file("r") as _f:
            f = _f.file_object
            points = f[points_location]
            geo = ET.SubElement(grid, "Geometry", GeometryType=geometry_type)
            dt, prec = numpy_to_xdmf_dtype[points.dtype.name]
            dim = "{} {}".format(*points.shape)
            data_item = ET.SubElement(
                geo,
                "DataItem",
                DataType=dt,
                Dimensions=dim,
                Format="HDF",
                Precision=prec,
            )
            data_item.text = f"{self.h5file}:/{points_location}"

    def _cells(self, grid: ET.Element, cells_location: str):
        with self._file("r") as _f:
            f = _f.file_object
            cells = f[cells_location]
            nb_cells = cells.shape[0]
            topo = ET.SubElement(
                grid,
                "Topology",
                TopologyType="Triangle",
                NumberOfElements=str(nb_cells),
            )
            dim = "{} {}".format(*cells.shape)
            dt, prec = numpy_to_xdmf_dtype[cells.dtype.name]
            data_item = ET.SubElement(
                topo,
                "DataItem",
                DataType=dt,
                Dimensions=dim,
                Format="HDF",
                Precision=prec,
            )
            data_item.text = f"{self.h5file}:/{cells_location}"

    def add_timeseries(self, steps: int, fields: list):
        collection = ET.SubElement(
            self.domain,
            "Grid",
            Name="TimeSeries",
            GridType="Collection",
            CollectionType="Temporal",
        )

        for i in range(steps):
            self.write_step(collection, fields, i)

    def write_step(self, collection: ET.Element, fields: list, step: int):
        """Write the data array for time t.

        Args:
            t (float): time
            data_location (str): String to data in h5 file
            name (str): Name for the field in the Xdmf file
        """
        with self._file("r") as _f:
            f = _f.file_object
            grid = ET.SubElement(collection, "Grid")
            ptr = f'xpointer(//Grid[@Name="{self.mesh_name}"]/*[self::Topology or self::Geometry])'

            ET.SubElement(
                grid, "{http://www.w3.org/2003/XInclude}include", xpointer=ptr
            )

            t = f[f"data/{fields[0]}/{step}"].attrs.get("t", step)
            ET.SubElement(grid, "Time", Value=str(t))

            for name in fields:
                self.write_attribute(grid, name, name, step)

    def write_attribute(
        self, grid: ET.Element, field_name: str, name: str, step: int
    ) -> None:
        """Write an attribute/field."""
        with self._file("r") as _f:
            f = _f.file_object
            data = f[f"data/{field_name}/{step}"]

            if data.ndim == 1 or data.shape[1] <= 1:
                att_type = "Scalar"
            elif data.ndim == 2:
                att_type = "Vector"
            elif data.ndim == 3 and len(set(data.shape[1:])) == 1:
                # Square shape -> Tensor
                att_type = "Tensor"
            else:
                att_type = "Matrix"

            # Cell or Node data
            center = data.attrs.get("center", "Node")

            att = ET.SubElement(
                grid,
                "Attribute",
                Name=name,
                AttributeType=att_type,
                Center=center,
            )

            dt, prec = numpy_to_xdmf_dtype[data.dtype.name]
            dim = " ".join([str(i) for i in data.shape])

            data_item = ET.SubElement(
                att,
                "DataItem",
                DataType=dt,
                Dimensions=dim,
                Format="HDF",
                Precision=prec,
            )
            data_item.text = f"{self.h5file}:/data/{field_name}/{step}"



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

from __future__ import annotations

import pandas as pd
import numpy as np

from bamboost.common.file_handler import FileHandler
from bamboost.common.hdf_pointer import Group

__all__ = ["GlobalGroup"]


class GlobalGroup(Group):
    """Enhanced Group for '/globals'.

    Args:
        file_handler: The file handler.
        path_to_data: The in-file path to the group.
    """

    def __init__(self, file_handler: FileHandler, path_to_data: str) -> None:
        super().__init__(file_handler, path_to_data)

    @property
    def df(self) -> pd.DataFrame:
        """Return a pandas DataFrame with all datasets."""
        d = {key: self[key][()] for key in self.datasets()}
        for key, val in d.items():
            d[key] = [i for i in val]
        return pd.DataFrame.from_dict(d)



__author__ = "florez@ethz.ch"
__copyright__ = ""
__license__ = "LGPLv3"

import logging
from typing import Literal

# Determine the version of the package
try:
    # If the package is installed, the version is stored in _version.py
    from bamboost._version import __version__
except ImportError:
    # This is necessary if the package is not installed
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)


def add_stream_handler(logger: logging.Logger) -> None:
    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter(
        "[%(asctime)s] %(name)s: %(levelname)s - %(message)s",
        style="%",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)


BAMBOOST_LOGGER = logging.getLogger("bamboost")
add_stream_handler(BAMBOOST_LOGGER)


def set_log_level(
    level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
) -> None:
    BAMBOOST_LOGGER.setLevel(level)


from bamboost._config import config  # noqa: E402, F401
from bamboost.extensions import extensions  # noqa: E402, F401
from bamboost.manager import Manager, ManagerCached  # noqa: E402, F401
from bamboost.simulation import Simulation  # noqa: E402, F401
from bamboost.simulation_writer import SimulationWriter  # noqa: E402, F401



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

from __future__ import annotations

import os
import pkgutil
import subprocess
from contextlib import contextmanager
from functools import cached_property
from typing import Any, Dict, Iterable, Tuple

import numpy as np
import pandas as pd
from typing_extensions import Self, deprecated

from bamboost import BAMBOOST_LOGGER, index
from bamboost._config import config, paths
from bamboost.accessors.fielddata import DataGroup
from bamboost.accessors.globals import GlobalGroup
from bamboost.accessors.meshes import Mesh, MeshGroup
from bamboost.common import hdf_pointer, utilities
from bamboost.common.file_handler import FileHandler, with_file_open
from bamboost.common.mpi import MPI
from bamboost.xdmf import XDMFWriter

__all__ = [
    "Simulation",
    "Links",
]

log = BAMBOOST_LOGGER.getChild(__name__.split(".")[-1])


class Links(hdf_pointer.MutableGroup):
    """Link group. Used to create and access links.

    I don't know how to distribute this to its own file in the accessors
    directory, due to circular imports.
    """

    def __init__(self, file_handler: FileHandler) -> None:
        super().__init__(file_handler, path_to_data="links")

    def _ipython_key_completions_(self):
        return tuple(self.all_links().keys())

    @with_file_open("r", driver="mpio")
    def __getitem__(self, key) -> Simulation:
        """Returns the linked simulation object."""
        return Simulation.fromUID(self.obj.attrs[key])

    def __setitem__(self, key, newvalue):
        """Creates the link."""
        return self.update_attrs({key: newvalue})

    def __delitem__(self, key):
        """Delete a link."""
        with self._file("a"):
            del self.obj.attrs[key]

    @with_file_open("r")
    def __repr__(self) -> str:
        return repr(
            pd.DataFrame.from_dict(self.all_links(), orient="index", columns=["UID"])
        )

    @with_file_open("r")
    def _repr_html_(self) -> str:
        return pd.DataFrame.from_dict(
            self.all_links(), orient="index", columns=["UID"]
        )._repr_html_()

    @with_file_open("r")
    def all_links(self) -> dict:
        return dict(self.obj.attrs)


class Simulation:
    """A single dataset/simulation. Used to write to it, read from it or append.

    Args:
        uid (str): unique identifier
        path (str): path to parent/database folder
        comm (MPI.Comm): MPI communicator (default=MPI.COMM_WORLD)
    """

    _mesh_location = "Mesh/0"
    _default_mesh = "mesh"

    def __init__(
        self,
        uid: str,
        path: str,
        comm: MPI.Comm = MPI.COMM_WORLD,
        create_if_not_exists: bool = False,
        *,
        _db_id: str = None,
    ):
        self.uid: str = uid
        path = comm.bcast(path, root=0)
        self.path_database: str = os.path.abspath(path)
        self.database_id = _db_id or index.get_uid_from_path(self.path_database)
        self.path: str = os.path.abspath(os.path.join(path, uid))
        self.h5file: str = os.path.join(self.path, f"{self.uid}.h5")
        self.xdmffile: str = os.path.join(self.path, f"{self.uid}.xdmf")

        if not os.path.exists(self.h5file) and not create_if_not_exists:
            raise FileNotFoundError(
                f"Simulation {self.uid} does not exist in {self.path}."
            )

        os.makedirs(self.path, exist_ok=True)

        # MPI information
        self._comm = comm
        self._psize = self._comm.size
        self._prank = self._comm.rank
        self._ranks = np.array([i for i in range(self._psize)])

        self._file = FileHandler(self.h5file)

        # Initialize groups to meshes, data and userdata. Create groups.
        self.meshes: MeshGroup = MeshGroup(self._file)
        self.data: DataGroup = DataGroup(self._file, self.meshes)
        self.globals: GlobalGroup = GlobalGroup(self._file, "/globals")
        self.userdata: hdf_pointer.MutableGroup = hdf_pointer.MutableGroup(
            self._file, "/userdata"
        )
        self.links: Links = Links(self._file)

    @classmethod
    def fromUID(
        cls, full_uid: str, *, index_database: index.IndexAPI = None, **kwargs
    ) -> Self:
        """Return the `Simulation` with given UID.

        Args:
            full_uid: the full id (Database uid : simulation uid)
        """
        if index_database is None:
            index_database = index.IndexAPI()
        db_uid, sim_uid = full_uid.split(":")
        db_path = index_database.get_path(db_uid)
        return cls(sim_uid, db_path, create_if_not_exists=False, **kwargs)

    @with_file_open()
    def __getitem__(self, key) -> hdf_pointer.BasePointer:
        """Direct access to HDF5 file.

        Returns:
            :class:`~bamboost.common.file_handler.BasePointer`
        """
        return hdf_pointer.BasePointer.new_pointer(self._file, key)

    def _repr_html_(self) -> str:
        html_string = pkgutil.get_data(__name__, "html/simulation.html").decode()
        icon = pkgutil.get_data(__name__, "html/icon.txt").decode()

        table_string = ""
        for key, value in self.parameters.items():
            if (
                isinstance(value, Iterable)
                and not isinstance(value, str)
                and len(value) > 5
            ):
                value = "..."
            table_string += f"""
            <tr>
                <td>{key}</td>
                <td>{value}</td>
            </tr>
            """

        metadata = self.metadata

        def get_pill_div(text: str, color: str):
            return (
                f'<div class="status" style="background-color:'
                f'var(--bb-{color});">{text}</div>'
            )

        status_options = {
            "Finished": get_pill_div("Finished", "green"),
            "Failed": get_pill_div("Failed", "red"),
            "Initiated": get_pill_div("Initiated", "grey"),
        }
        submitted_options = {
            True: get_pill_div("Submitted", "green"),
            False: get_pill_div("Not submitted", "grey"),
        }

        html_string = (
            html_string.replace("$UID", self.uid)
            .replace("$ICON", icon)
            .replace("$TREE", self.show_files(printit=False).replace("\n", "<br>"))
            .replace("$TABLE", table_string)
            .replace("$NOTE", metadata["notes"])
            .replace(
                "$STATUS",
                status_options.get(
                    metadata["status"],
                    f'<div class="status">{metadata["status"]}</div>',
                ),
            )
            .replace("$SUBMITTED", submitted_options[metadata.get("submitted", False)])
            .replace("$TIMESTAMP", metadata["time_stamp"])
        )
        return html_string

    def _push_update_to_sqlite(self, update_dict: dict) -> None:
        """Push update to sqlite database.

        Args:
            - update_dict (dict): key value pair to push
        """
        if not config["options"].get("sync_table", True):
            return
        try:
            index.DatabaseTable(self.database_id).update_entry(self.uid, update_dict)
        except index.Error as e:
            log.warning(f"Could not update sqlite database: {e}")

    @cached_property
    def parameters(self) -> Dict[str, Any]:
        tmp_dict = dict()
        if self._prank == 0:
            with self._file("r"):
                # return if parameters is not in the file
                if "parameters" not in self._file.keys():
                    return {}
                tmp_dict.update(self._file["parameters"].attrs)
                for key in self._file["parameters"].keys():
                    tmp_dict.update({key: self._file[f"parameters/{key}"][()]})

        tmp_dict = utilities.unflatten_dict(tmp_dict)

        tmp_dict = self._comm.bcast(tmp_dict, root=0)
        return tmp_dict

    @property
    def metadata(self) -> dict:
        tmp_dict = dict()
        if self._prank == 0:
            with self._file("r") as file:
                tmp_dict.update(file.attrs)

        tmp_dict = utilities.unflatten_dict(tmp_dict)
        tmp_dict = self._comm.bcast(tmp_dict, root=0)
        return tmp_dict

    def files(self, filename: str) -> str:
        """Get the path to the file.

        Args:
            filename: name of the file
        """
        return os.path.join(self.path, filename)

    def show_files(
        self, level=-1, limit_to_directories=False, length_limit=1000, printit=True
    ) -> str:
        """Show the file tree of the simulation directory.

        Args:
            level: how deep to print the tree
            limit_to_directories: only print directories
            length_limit: cutoff
        """
        tree_string = utilities.tree(
            self.path, level, limit_to_directories, length_limit
        )
        if printit:
            print(tree_string)
        else:
            return tree_string

    def open_in_file_explorer(self) -> None:
        """Open the simulation directory. Uses `xdg-open` on linux systems."""
        if os.name == "nt":  # should work on Windows
            os.startfile(self.path)
        else:
            subprocess.run(["xdg-open", self.path])

    def open_in_paraview(self) -> None:
        """Open the xdmf file in paraview."""
        subprocess.Popen(["paraview", self.xdmffile])

    def get_full_uid(self) -> str:
        """Returns the full uid of the simulation (including the one of the database)"""
        database_uid = index.get_uid_from_path(self.path_database)
        return f"{database_uid}:{self.uid}"

    def change_status(self, status: str) -> None:
        """Change status of simulation.

        Args:
            status (str): new status
        """
        if self._prank == 0:
            self._file.open("a")
            self._file.attrs["status"] = status
            self._file.close()

        self._push_update_to_sqlite({"status": status})

    def update_metadata(self, update_dict: dict) -> None:
        """Update the metadata attributes.

        Args:
            update_dict: dictionary to push
        """
        if self._prank == 0:
            update_dict = utilities.flatten_dict(update_dict)
            with self._file("a") as file:
                file.attrs.update(update_dict)

            self._push_update_to_sqlite(update_dict)

    def update_parameters(self, update_dict: dict) -> None:
        """Update the parameters dictionary.

        Args:
            update_dict: dictionary to push
        """
        if self._prank == 0:
            update_dict = utilities.flatten_dict(update_dict)
            with self._file("a") as file:
                file["parameters"].attrs.update(update_dict)

            self._push_update_to_sqlite(update_dict)

    def create_xdmf_file(self, fields: list = None, nb_steps: int = None) -> None:
        """Create the xdmf file to read in paraview.

        Args:
            fields (list[str]): fields for which to write timeseries information,
                if not specified, all fields in data are written.
            nb_steps (int): number of steps the simulation has
        """

        if self._prank == 0:
            with self._file("r"):
                f = self._file.file_object
                if "data" not in f.keys():
                    fields, nb_steps = [], 0
                if fields is None:
                    fields = list(f["data"].keys())

                if nb_steps is None:
                    grp_name = list(f["data"].keys())[0]
                    nb_steps = list(f[f"data/{grp_name}"].keys())
                    nb_steps = max(
                        [
                            int(step)
                            for step in nb_steps
                            if not (
                                step.startswith("__") or step.endswith("_intermediates")
                            )
                        ]
                    )

                # temporary fix to load coordinates/geometry
                coords_name = (
                    "geometry"
                    if "geometry"
                    in f[f"{self._mesh_location}/{self._default_mesh}"].keys()
                    else "coordinates"
                )

            with self._file("r"):
                xdmf_writer = XDMFWriter(self.xdmffile, self._file)
                xdmf_writer.write_points_cells(
                    f"{self._mesh_location}/{self._default_mesh}/{coords_name}",
                    f"{self._mesh_location}/{self._default_mesh}/topology",
                )

                if fields:
                    xdmf_writer.add_timeseries(nb_steps + 1, fields)
                xdmf_writer.write_file()

        self._comm.barrier()

    def create_run_script(
        self,
        commands: list[str],
        euler: bool = True,
        sbatch_kwargs: dict[str, Any] = None,
    ) -> None:
        """Create a batch job and put it into the folder.

        Args:
            commands: A list of strings being the user defined commands to run
            euler: If false, a local bash script will be written
            sbatch_kwargs: Additional sbatch arguments.
                This parameter allows you to provide additional arguments to the `sbatch` command
                when submitting jobs to a Slurm workload manager. The arguments should be provided
                in the format of a dict of sbatch option name and values.

                Use this parameter to specify various job submission options such as the number of
                tasks, CPU cores, memory requirements, email notifications, and other sbatch options
                that are not covered by default settings.
                By default, the following sbatch options are set:
                - `--output`: The output file is set to `<uid>.out`.
                - `--job-name`: The job name is set to `<full_uid>`.

                The following arguments should bring you far:
                - `--ntasks`: The number of tasks to run. This is the number of MPI processes to start.
                - `--mem-per-cpu`: The memory required per CPU core.
                - `--time`: The maximum time the job is allowed to run.
                - `--tmp`: Temporary scratch space to use for the job.
        """

        def _set_environment_variables():
            return (
                f"""DATABASE_DIR=$(sqlite3 {paths['DATABASE_FILE']} "SELECT path FROM dbindex WHERE id='{self.database_id}'")\n"""
                f"export SIMULATION_DIR=$DATABASE_DIR/{self.uid}\n"
                f"export SIMULATION_ID={self.database_id}:{self.uid}\n\n"
            )

        script = "#!/bin/bash\n\n"
        if euler:
            # Write sbatch submission script for EULER
            # filename: sbatch_{self.uid}.sh
            if sbatch_kwargs is None:
                sbatch_kwargs = {}

            sbatch_kwargs.setdefault("--output", f"{self.path}/{self.uid}.out")
            sbatch_kwargs.setdefault("--job-name", self.get_full_uid())

            for key, value in sbatch_kwargs.items():
                script += f"#SBATCH {key}={value}\n"

            script += "\n"
            script += _set_environment_variables()
            script += "\n".join(commands)

            with open(self.files(f"sbatch_{self.uid}.sh"), "w") as file:
                file.write(script)
        else:
            # Write local bash script
            # filename: {self.uid}.sh
            script += _set_environment_variables()
            script += "\n".join(commands)

            with open(self.files(f"{self.uid}.sh"), "w") as file:
                file.write(script)

        with self._file("a") as file:
            file.attrs.update({"submitted": False})
        self._push_update_to_sqlite({"submitted": False})

    @deprecated("use `create_run_script` instead")
    def create_batch_script(self, *args, **kwargs):
        return self.create_run_script(*args, **kwargs)

    def submit(self) -> None:
        """Submit the job for this simulation."""
        if f"sbatch_{self.uid}.sh" in os.listdir(self.path):
            batch_script = os.path.abspath(
                os.path.join(self.path, f"sbatch_{self.uid}.sh")
            )
            env = os.environ.copy()
            _ = env.pop("BAMBOOST_MPI", None)
            subprocess.run(["sbatch", f"{batch_script}"], env=env)
        elif f"{self.uid}.sh" in os.listdir(self.path):
            bash_script = os.path.abspath(os.path.join(self.path, f"{self.uid}.sh"))
            env = os.environ.copy()
            _ = env.pop("BAMBOOST_MPI", None)
            subprocess.run(["bash", f"{bash_script}"], env=env)
        else:
            raise FileNotFoundError(
                f"Could not find a batch script for simulation {self.uid}."
            )

        log.info(f"Simulation {self.uid} submitted!")

        with self._file("a") as file:
            file.attrs.update({"submitted": True})

        self._push_update_to_sqlite({"submitted": True})

    @with_file_open("a")
    def change_note(self, note) -> None:
        if self._prank == 0:
            self._file.attrs["notes"] = note
            self._push_update_to_sqlite({"notes": note})

    # Ex-Simulation reader methods
    # ----------------------------

    def open(self, mode: str = "r", driver=None, comm=None) -> FileHandler:
        """Use this as a context manager in a `with` statement.
        Purpose: keeping the file open to directly access/edit something in the
        HDF5 file of this simulation.

        Args:
            mode (`str`): file mode (see h5py docs)
            driver (`str`): file driver (see h5py docs)
            comm (`str`): mpi communicator
        """
        return self._file(mode, driver, comm)

    @property
    def mesh(self) -> Mesh:
        """Return the default mesh.

        Returns:
            MeshGroup
        """
        return self.meshes[self._default_mesh]

    @with_file_open("r")
    def get_mesh(self, mesh_name: str = None) -> Tuple[np.ndarray, np.ndarray]:
        """Return coordinates and connectivity. Currently returns numpy arrays.

        Args:
            mesh_name (`str`): optional, name of mesh to read (default = mesh)
        Returns:
            Tuple of np.arrays (coordinates, connectivity)
        """
        if mesh_name is None:
            mesh_name = self._default_mesh

        # Raise an error if the mesh is not found
        if (self._mesh_location.split("/")[0] not in self._file.keys()) or (
            mesh_name not in self._file[self._mesh_location].keys()
        ):
            raise KeyError(f"Mesh location {self._mesh_location} not found in file.")

        mesh = self.meshes[mesh_name]
        return mesh.coordinates, mesh.connectivity

    @property
    @deprecated("Use `data.info` instead")
    @with_file_open("r")
    def data_info(self) -> pd.DataFrame:
        """View the data stored.

        Returns:
            :class:`pd.DataFrame`
        """
        tmp_dictionary = dict()
        for data in self.data:
            steps = len(data)
            shape = data.obj["0"].shape
            dtype = data.obj["0"].dtype
            tmp_dictionary[data._name] = {
                "dtype": dtype,
                "shape": shape,
                "steps": steps,
            }
        return pd.DataFrame.from_dict(tmp_dictionary)

    @property
    @with_file_open("r")
    def git(self) -> dict:
        """Get Git information.

        Returns:
            :class:`dict` with different repositories.
        """
        if "git" not in self._file.keys():
            return "Sorrrry, no git information stored :()"
        grp = self._file["git"]
        tmp_dict = {}
        for repo in grp.keys():
            tmp_dict[repo] = grp[repo][()].decode("utf8")
        return tmp_dict

    def get_data_interpolator(self, field: str, step: int):
        """Get Linear interpolator for data field at step. Uses the linked mesh.

        Args:
            name (`str`): name of the data field
            step (`int`): step
        Returns:
            :class:`scipy.interpolate.LinearNDInterpolator`
        """
        from scipy.interpolate import LinearNDInterpolator

        return LinearNDInterpolator(
            self.data[field].mesh.coordinates, self.data[field].at_step(step)
        )

    @with_file_open()
    def show_h5tree(self) -> None:
        """Print the tree inside the h5 file."""
        # print('\U00002B57 ' + os.path.basename(self.h5file))
        print("\U0001f43c " + os.path.basename(self.h5file))
        utilities.h5_tree(self._file.file_object)

    @contextmanager
    def enter_path(self):
        """A context manager for changing the working directory to this simulations' path.

        >>> with sim.working_directory():
        >>>     ...
        """

        current_dir = os.getcwd()
        try:
            os.chdir(self.path)
            yield
        finally:
            os.chdir(current_dir)



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code
"""
This module provides a class to handle sqlite databases.
"""

from __future__ import annotations

import json
import sqlite3
from contextlib import contextmanager
from functools import wraps
from typing import Generator, Iterable

import numpy as np
from typing_extensions import Self

from bamboost import BAMBOOST_LOGGER
from bamboost._config import config
from bamboost.common.mpi import MPI

__all__ = ["SQLiteHandler"]

log = BAMBOOST_LOGGER.getChild(__name__.split(".")[-1])

_type_to_sql_column_type = {
    np.ndarray: "ARRAY",
    np.datetime64: "DATETIME",
    int: "INTEGER",
    float: "REAL",
    str: "TEXT",
    bool: "BOOL",
}


def get_sqlite_column_type(val):
    dtype = type(val)
    if isinstance(val, np.generic):
        dtype = type(val.item())
    if dtype in _type_to_sql_column_type:
        return _type_to_sql_column_type[dtype]

    if isinstance(val, Iterable):
        return "JSON"
    if isinstance(val, dict):
        return "JSON"
    if isinstance(val, np.ndarray):
        return "ARRAY"


def _register_sqlite_adapters():
    # Converts np.array & Iterable to JSON when inserting
    sqlite3.register_adapter(np.ndarray, lambda arr: json.dumps(arr.tolist()))
    sqlite3.register_adapter(list, lambda val: json.dumps(val))
    sqlite3.register_adapter(dict, lambda val: json.dumps(val))
    sqlite3.register_adapter(tuple, lambda val: json.dumps(val))
    sqlite3.register_adapter(set, lambda val: json.dumps(val))

    # Numpy generic types
    def adapt_numpy_number(val):
        return val.item()

    sqlite3.register_adapter(np.int_, adapt_numpy_number)
    sqlite3.register_adapter(np.float64, adapt_numpy_number)
    sqlite3.register_adapter(np.datetime64, adapt_numpy_number)
    sqlite3.register_adapter(np.bool_, bool)


def _register_sqlite_converters(convert_arrays: bool = True):
    # Converts JSON to np.array & Iterable when selecting
    if convert_arrays:
        sqlite3.register_converter("ARRAY", lambda text: np.array(json.loads(text)))
    else:
        sqlite3.register_converter("ARRAY", lambda text: json.loads(text))
    sqlite3.register_converter("JSON", lambda text: json.loads(text))

    def convert_bool(val):
        # from now on, all bools should be stored as integers
        try:
            return bool(int(val))
        # for a while, booleans were stored as bytes
        # this exception handles the conversion of these bytes to bool for old databases
        except ValueError:
            return bool(int.from_bytes(val, "big"))

    sqlite3.register_converter("BOOL", convert_bool)


# ----------------
# DECORATORS
# ----------------
def with_connection(func):
    """Decorator to ensure that the cursor is available. If the cursor is not
    available, the connection is opened and closed after the function is
    executed."""

    @wraps(func)
    def wrapper(self: SQLiteHandler, *args, **kwargs):
        cursor: sqlite3.Cursor = self._cursor
        # check if cursor is available
        if not self._is_open or cursor is None:
            with self.open():
                return func(self, *args, **kwargs)
        return func(self, *args, **kwargs)

    return wrapper


class SQLiteHandler:
    """Class to handle sqlite databases."""

    def __init__(
        self, file: str, _comm=MPI.COMM_WORLD, convert_arrays: bool = True
    ) -> None:
        if _comm.rank != 0:
            return
        # self._comm = _comm
        self.file = file
        self._conn = None
        self._cursor = None
        self._is_open = False
        self._lock_stack = 0

        _register_sqlite_adapters()
        _register_sqlite_converters(convert_arrays=convert_arrays)

    @property
    def _lock_stack(self) -> int:
        return self.__lock_stack

    @_lock_stack.setter
    def _lock_stack(self, value: int) -> None:
        self.__lock_stack = value if value >= 0 else 0

    def connect(self) -> Self:
        self._lock_stack += 1
        if not self._is_open:
            log.debug(f"Connecting to {self.file}")
            self._conn = sqlite3.connect(
                self.file, detect_types=sqlite3.PARSE_DECLTYPES
            )
            self._cursor = self._conn.cursor()
            self._is_open = True
        return self

    def close(self, *, force: bool = False, ensure_commit: bool = False) -> Self:
        if force:
            self._lock_stack = 0
        else:
            self._lock_stack -= 1

        if self._lock_stack <= 0 and self._is_open:
            log.debug(f"Closing connection to {self.file}")
            self._conn.commit()
            self._conn.close()
            self._is_open = False
            return self

        if ensure_commit:
            self._conn.commit()

        return self

    def commit(self) -> None:
        self._conn.commit()
        return self

    @contextmanager
    def open(
        self,
        *,
        force_close: bool = False,
        ensure_commit: bool = False,
    ) -> Generator[Self]:
        """The open method is used as a context manager.

        Args:
            - ensure_commit (bool, optional): Ensure that the connection is
              committed. Defaults to False.

        Example:
            >>> with index.open() as table:
            >>>     table._cursor.execute("SELECT * FROM database")
        """
        self.connect()
        try:
            yield self
        finally:
            self.close(ensure_commit=ensure_commit, force=force_close)



import argparse
import time

import h5py
import numpy as np
from mpi4py import MPI

from bamboost import SimulationWriter


def run_test(path: str, uid: str) -> None:
    """run the test."""
    sim = SimulationWriter(uid, path)

    array_size = sim.parameters["array_size"]
    nb_processes = sim.parameters["nb_processes"]
    nb_steps = sim.parameters["nb_steps"]
    array_size_per_process = array_size[0] // nb_processes
    dim = array_size[1]

    # Start time
    sim._comm.Barrier()
    start = MPI.Wtime()

    with sim:
        for _ in range(nb_steps):
            sim.add_field("test", np.random.randn(array_size_per_process, dim))
            sim.finish_step()

    # End time
    sim._comm.Barrier()
    end = MPI.Wtime()

    # Store time
    sim.userdata["write_time_total"] = end - start

    # Print result:
    if sim._comm.rank == 0:
        print(
            f"Writing {nb_steps} steps of {array_size_per_process}x{dim} "
            f"arrays with {nb_processes} processes took {end - start:.3f} seconds."
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--path", type=str)
    parser.add_argument("--uid", type=str)
    args = parser.parse_args()
    run_test(args.path, args.uid)




import numpy as np
import argparse

from dbmanager.simulation import SimulationWriter

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--path", type=str)
    parser.add_argument("--uid", type=str)
    args = parser.parse_args()
    uid, path = args.uid, args.path

    # Use with statement for keeping track of the status of the simulation
    with SimulationWriter(uid, path) as writer:
        writer.register_git_attributes()
        writer.add_metadata()

        # to load the parameters of this simulation use
        params = writer.parameters

        # do what you have to do
        # here we create a simple mesh
        coords = np.array([[0.0, 0.0], [1.0, 0.0], [2.0, 0.0], [0.0, 1.0], [1.0, 2.0]])

        initial_values = np.random.randint(0, 10, 5)

        for t in range(params["timesteps"]):
            result = initial_values * t ** params["exponent"]

            # write the coordinates
            # if you have a mesh with coords and connectivity, you can add a global mesh with
            # writer.add_mesh(coords, conn) outside this for loop
            writer.add_field("coords", coords, time=t)

            # write the data
            writer.add_field("mult_with_t", result, time=t)

            # write a global quantity
            writer.add_global_field("sum", np.sum(result))

            # finish the step, required
            writer.finish_step()
